
from escnn.nn import GeometricTensor
from .equivariant_module import EquivariantModule

import torch

from typing import List, Tuple, Union, Any, Iterator

from collections import OrderedDict

__all__ = ["SequentialModule"]


class SequentialModule(EquivariantModule):
    
    def __init__(self,
                 *args: EquivariantModule,
                 ):
        r"""
        
        A sequential container similar to :class:`torch.nn.Sequential`.
        
        The constructor accepts both a list or an ordered dict of :class:`~escnn.nn.EquivariantModule` instances.

        The module also supports indexing, slicing and iteration.
        If slicing with a step different from 1 is used, one should ensure that adjacent modules in the new sequence
        are compatible.
        
        Example::
        
            # Example of SequentialModule
            s = escnn.gspaces.rot2dOnR2(8)
            c_in = escnn.nn.FieldType(s, [s.trivial_repr]*3)
            c_out = escnn.nn.FieldType(s, [s.regular_repr]*16)
            model = escnn.nn.SequentialModule(
                      escnn.nn.R2Conv(c_in, c_out, 5),
                      escnn.nn.InnerBatchNorm(c_out),
                      escnn.nn.ReLU(c_out),
            )

            len(module) # returns 3

            module[:2] # returns another SequentialModule containing the first two modules

            # Example with OrderedDict
            s = escnn.gspaces.rot2dOnR2(8)
            c_in = escnn.nn.FieldType(s, [s.trivial_repr]*3)
            c_out = escnn.nn.FieldType(s, [s.regular_repr]*16)
            model = escnn.nn.SequentialModule(OrderedDict([
                      ('conv', escnn.nn.R2Conv(c_in, c_out, 5)),
                      ('bn', escnn.nn.InnerBatchNorm(c_out)),
                      ('relu', escnn.nn.ReLU(c_out)),
            ]))
        
        """
        
        super(SequentialModule, self).__init__()

        self.in_type = None
        self.out_type = None
        
        if len(args) == 1 and isinstance(args[0], OrderedDict):
            for key, module in args[0].items():
                assert isinstance(module, EquivariantModule)
                self.add_module(key, module)
        else:
            for idx, module in enumerate(args):
                assert isinstance(module, EquivariantModule)
                self.add_module(str(idx), module)
        
        # for i in range(1, len(self._modules.values())):
        #     assert self._modules.values()[i-1].out_type == self._modules.values()[i].in_type
        
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        
        Args:
            input (GeometricTensor): the input GeometricTensor

        Returns:
            the output tensor
            
        """
        
        assert input.type == self.in_type
        x = input
        for m in self._modules.values():
            x = m(x)

        assert x.type == self.out_type
        
        return x
    
    def add_module(self, name: str, module: EquivariantModule):
        r"""
        Append ``module`` to the sequence of modules applied in the forward pass.
        
        """
        
        if len(self._modules) == 0:
            assert self.in_type is None
            assert self.out_type is None
            self.in_type = module.in_type
        else:
            assert module.in_type == self.out_type, f"{module.in_type} != {self.out_type}"
            
        self.out_type = module.out_type
        super(SequentialModule, self).add_module(name, module)

    def append(self, module: EquivariantModule) -> 'SequentialModule':
        r"""Appends a new EquivariantModule at the end.
        """
        self.add_module(str(len(self)), module)
        return self

    def __getitem__(self, idx) -> Union['SequentialModule', EquivariantModule]:
        if isinstance(idx, slice):
            return self.__class__(OrderedDict(list(self._modules.items())[idx]))
        elif isinstance(idx, int):
            assert -len(self) <= idx < len(self), (idx, len(self))
            idx = idx % len(self)
            for i, module in enumerate(self._modules.values()):
                if i == idx:
                    return module
            raise ValueError(f'Index {idx} not found!')
        else:
            raise ValueError(f'Index {idx} not valid!')

    def __iter__(self) -> Iterator[EquivariantModule]:
        return iter(self._modules.values())

    def __len__(self) -> int:
        return len(self._modules)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        
        assert len(input_shape) > 1
        assert input_shape[1] == self.in_type.size
        
        out_shape = input_shape

        for m in self._modules.values():
            out_shape = m.evaluate_output_shape(out_shape)
        
        return out_shape

    def check_equivariance(self, atol: float = 2e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        return super(SequentialModule, self).check_equivariance(atol=atol, rtol=rtol)

    def export(self):
        r"""
        Export this module to a normal PyTorch :class:`torch.nn.Sequential` module and set to "eval" mode.

        """
    
        self.eval()
    
        submodules = []
        
        # convert all the submodules if necessary
        for name, module in self._modules.items():
            if isinstance(module, EquivariantModule):
                module = module.export()
                
            submodules.append(
                (name, module)
            )

        return torch.nn.Sequential(OrderedDict(submodules))




from .basis import KernelBasis, EmptyBasisException
from .steerable_filters_basis import SteerableFiltersBasis

from escnn.group import Group
from escnn.group import IrreducibleRepresentation
from escnn.group import Representation

import torch

from typing import Type, Union, Tuple, Dict, List, Iterable, Callable, Set
from abc import ABC, abstractmethod
from collections import defaultdict


class IrrepBasis(KernelBasis):
    
    def __init__(self,
                 basis: SteerableFiltersBasis,
                 in_irrep: Union[IrreducibleRepresentation, Tuple],
                 out_irrep: Union[IrreducibleRepresentation, Tuple],
                 dim: int,
                 harmonics: List[Tuple] = None
                 ):
        r"""

        Abstract class for bases implementing the kernel constraint solutions associated to irreducible input and output
        representations.

        .. note ::
            The steerable *filter* ``basis`` is not necessarily associated with the same group as ``in_irrep`` and
            ``out_irrep``.
            For instance, :class:`~escnn.kernels.RestrictedWignerEckartBasis` uses a larger group to define ``basis``.
            The attribute ``IrrepBasis.group``, instead, refers to the equivariance group of this steerable *kernel*
            basis and is the same group of ``in_irrep`` and ``out_irrep``.
            The irreps in the list ``harmonics`` refer to the group in the steerable filter ``basis``,
            and not to ``IrrepBasis.group``.

        Args:
            basis (SteerableFiltersBasis): the steerable basis used to parameterize scalar filters and generate the kernel solutions
            in_irrep (IrreducibleRepresentation): the input irrep
            out_irrep (IrreducibleRepresentation): the output irrep
            dim (int): the number of elements in the basis
            harmonics (list, optional): optionally, use only a subset of the steerable filters in ``basis``. This list
                                        defines a subset of the group's irreps and is used to select only the steerable
                                        basis filters which transform according to these irreps.

        Attributes:
            ~.group (Group): the equivariance group
            ~.in_irrep (IrreducibleRepresentation): the input irrep
            ~.out_irrep (IrreducibleRepresentation): the output irrep
            ~.basis (SteerableFiltersBasis): the steerable basis used to parameterize scalar filters

        """

        super(IrrepBasis, self).__init__(dim, (out_irrep.size, in_irrep.size))

        assert in_irrep.group == out_irrep.group
        self.group: Group = in_irrep.group
        self.in_irrep: IrreducibleRepresentation = self.group.irrep(*self.group.get_irrep_id(in_irrep))
        self.out_irrep: IrreducibleRepresentation = self.group.irrep(*self.group.get_irrep_id(out_irrep))
        self.basis: SteerableFiltersBasis = basis

        self.js = []
        harmonics = set(harmonics)
        for j, _ in basis.js:
            if j not in self.js and (harmonics is None or j in harmonics):
                self.js.append(j)

        self._start_index = {}
        idx = 0
        for _j in self.js:
            self._start_index[_j] = idx
            idx += self.dim_harmonic(_j)

    def sample(self, points: torch.Tensor, out: torch.Tensor = None) -> torch.Tensor:
        r"""

        Sample the continuous basis elements on the discrete set of points in ``points``.
        Optionally, store the resulting multidimentional array in ``out``.

        ``points`` must be an array of shape `(N, d)`, where `N` is the number of points and `d` is the dimensionality
        of the Euclidean space where filters are defined.

        Args:
            points (~torch.Tensor): points where to evaluate the basis elements
            out (~torch.Tensor, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """
    
        assert len(points.shape) == 2
        S = points.shape[0]
    
        if out is None:
            out = torch.empty(S, self.dim, self.shape[0], self.shape[1], device=points.device, dtype=points.dtype)
    
        assert out.shape == (S, self.dim, self.shape[0], self.shape[1])

        steerable_basis = self.basis.sample_as_dict(points)

        B = 0
        outs = {}
        for b, j in enumerate(self.js):
            outs[j] = out[:, B:B + self.dim_harmonic(j), ...]
            B += self.dim_harmonic(j)

        self.sample_harmonics(steerable_basis, outs)
        return out

    @abstractmethod
    def sample_harmonics(self, points: Dict[Tuple, torch.Tensor], out: Dict[Tuple, torch.Tensor] = None) -> Dict[Tuple, torch.Tensor]:
        r"""

        Sample the continuous basis elements on the discrete set of points.
        Rather than using the points' coordinates, the method directly takes in input the steerable basis elements
        sampled on this points using the method :meth:`escnn.kernels.SteerableFilterBasis.sample_as_dict` of
        ``self.basis``.

        Similarly, rather than returning a single tensor containing all sampled basis elements, it groups basis elements
        by the ``G``-irrep acting on them.
        The method returns a dictionary mapping each irrep's ``id`` to a tensor of shape `(N, m, o, i)`, where
        `N` is the number of points,
        `m` is the multiplicity of the irrep (see :meth:`~escnn.kernels.SteerableKernelBasis.dim_harmonic`)
        and `o, i` is the number of input and output channels (see the ``shape`` attribute).

        Optionally, store the resulting tensors in ``out``, rather than allocating new memory.

        Args:
            points (~torch.Tensor): points where to evaluate the basis elements
            out (~torch.Tensor, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """
        raise NotImplementedError()

    @abstractmethod
    def dim_harmonic(self, j: Tuple) -> int:
        r'''
        Number of kernel basis elements generated from elements of the steerable filter basis (``self.basis``) which
        transform according to the ``self.basis.group``-irrep identified by ``j``.
        '''
        raise NotImplementedError()
    
    @abstractmethod
    def attrs_j_iter(self, j: Tuple) -> Iterable:
        raise NotImplementedError()
    
    @abstractmethod
    def attrs_j(self, j: Tuple, idx) -> Dict:
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def _generator(cls,
                   basis: SteerableFiltersBasis,
                   psi_in: Union[IrreducibleRepresentation, Tuple],
                   psi_out: Union[IrreducibleRepresentation, Tuple],
                   **kwargs
    ) -> 'IrrepBasis':
        raise NotImplementedError()


class SteerableKernelBasis(KernelBasis):
    
    def __init__(self,
                 basis: SteerableFiltersBasis,
                 in_repr: Representation,
                 out_repr: Representation,
                 irreps_basis: Type[IrrepBasis],
                 **kwargs):
        r"""
        
        Implements a general basis for the vector space of equivariant kernels over an Euclidean space :math:`X=\R^n`.
        A :math:`G`-equivariant kernel :math:`\kappa`, mapping between an input field, transforming under
        :math:`\rho_\text{in}` (``in_repr``), and an output field, transforming under  :math:`\rho_\text{out}`
        (``out_repr``), satisfies the following constraint:
        
        .. math ::
            
            \kappa(gx) = \rho_\text{out}(g) \kappa(x) \rho_\text{in}(g)^{-1} \qquad \forall g \in G, \forall x \in X
        
        As the kernel constraint is a linear constraint, the space of equivariant kernels is a vector subspace of the
        space of all convolutional kernels. It follows that any equivariant kernel can be expressed in terms of a basis
        of this space.
        
        This class solves the kernel constraint for two arbitrary representations by combining the solutions of the
        kernel constraints associated to their :class:`~escnn.group.IrreducibleRepresentation` s.
        In order to do so, it relies on ``irreps_basis`` which solves individual irreps constraints. ``irreps_basis``
        must be a class (subclass of :class:`~escnn.kernels.IrrepsBasis`) which builds a basis for equivariant
        kernels associated with irreducible representations when instantiated.
        
        The groups :math:`G` which are currently implemented are origin-preserving isometries (what are called
        structure groups, or sometimes gauge groups, in the language of
        `Gauge Equivariant CNNs <https://arxiv.org/abs/1902.04615>`_ ).
        The origin-preserving isometries of :math:`\R^d` are subgroups of :math:`O(d)`, i.e. reflections and rotations.
        Therefore, equivariance does not enforce any constraint on the radial component of the kernels.
        Hence, this class only implements a basis for the angular part of the kernels.
        
        .. warning ::
            
            Typically, the user does not need to manually instantiate this class.
            Instead, we suggest to use the interface provided in :doc:`escnn.gspaces`.
        
        Args:
            basis (SteerableFiltersBasis): a steerable basis for scalar filters over the base space
            in_repr (Representation): Representation associated with the input feature field
            out_repr (Representation): Representation associated with the output feature field
            irreps_basis (class): class defining the irreps basis. This class is instantiated for each pair of irreps to solve all irreps constraints.
            **kwargs: additional arguments used when instantiating ``irreps_basis``

        Attributes:
            ~.group (Group): the equivariance group ``G``.
            ~.in_repr (Representation): the input representation
            ~.out_repr (Representation): the output representation

        """
        
        assert in_repr.group == out_repr.group
        
        self.in_repr: Representation = in_repr
        self.out_repr: Representation = out_repr
        group = in_repr.group
        self.group: Group = group
        
        self._irrep_basis = irreps_basis
        self._irrep__basis_kwargs = kwargs

        ################

        # Dict[Tuple, IrrepsBasis]:
        self.irreps_bases = {}

        js = set()

        # loop over all input irreps
        for i_irrep_id in set(in_repr.irreps):
            # loop over all output irreps
            for o_irrep_id in set(out_repr.irreps):
                try:
                    # retrieve the irrep intertwiner basis
                    intertwiner_basis = irreps_basis._generator(basis, i_irrep_id, o_irrep_id, **kwargs)
                    assert intertwiner_basis.group == self.group
                    self.irreps_bases[(i_irrep_id, o_irrep_id)] = intertwiner_basis
                    # compute the set of all harmonics needed by all intertwiners bases
                    # in this way, we can precompute the embedding of the points with the harmonics and reuse the same
                    # embeddings for all irreps bases
                    js.update(intertwiner_basis.js)
                except EmptyBasisException:
                    # if the basis is empty, skip it
                    pass

        self._dim_harmonics = defaultdict(int)
        self.bases = [[None for _ in range(len(out_repr.irreps))] for _ in range(len(in_repr.irreps))]
        dim = 0
        # loop over all input irreps
        for ii, i_irrep_id in enumerate(in_repr.irreps):
            # loop over all output irreps
            for oo, o_irrep_id in enumerate(out_repr.irreps):
                if (i_irrep_id, o_irrep_id) in self.irreps_bases:
                    self.bases[ii][oo] = self.irreps_bases[(i_irrep_id, o_irrep_id)]
                    dim += self.irreps_bases[(i_irrep_id, o_irrep_id)].dim
                    for j in self.bases[ii][oo].js:
                        self._dim_harmonics[j] += self.irreps_bases[(i_irrep_id, o_irrep_id)].dim_harmonic(j)

        ################

        # before registering tensors as buffers and sub-modules, we need to call torch.nn.Module.__init__()
        super(SteerableKernelBasis, self).__init__(dim, (out_repr.size, in_repr.size))

        self.basis = basis

        for io_pair, intertwiner_basis in self.irreps_bases.items():
            self.add_module(f'basis_{io_pair}', intertwiner_basis)

        ################

        A_inv = torch.tensor(in_repr.change_of_basis_inv, dtype=torch.float32).clone()
        B = torch.tensor(out_repr.change_of_basis, dtype=torch.float32).clone()

        if not torch.allclose(A_inv, torch.eye(in_repr.size)):
            self.register_buffer('A_inv', A_inv)
        else:
            self.A_inv = None

        if not torch.allclose(B, torch.eye(out_repr.size)):
            self.register_buffer('B', B)
        else:
            self.B = None

        self.js = [j for j, m in self.basis.js if j in js]

        if self.basis.group.trivial_representation.id in self.js:
            # make sure that the harmonic corresponding to the trivial representation is the first in the list.
            self.js.remove(self.basis.group.trivial_representation.id)
            self.js = [self.basis.group.trivial_representation.id] + self.js
        
        self.in_sizes = []
        self.out_sizes = []
        # loop over all input irreps
        for ii, i_irrep_id in enumerate(in_repr.irreps):
            self.in_sizes.append(group.irrep(*i_irrep_id).size)
            
        # loop over all output irreps
        for oo, o_irrep_id in enumerate(out_repr.irreps):
            self.out_sizes.append(group.irrep(*o_irrep_id).size)

        self._slices = defaultdict(dict)
        basis_count = defaultdict(int)
        in_position = 0
        for ii, in_size in enumerate(self.in_sizes):
            out_position = 0
            for oo, out_size in enumerate(self.out_sizes):
                if self.bases[ii][oo] is not None:
                    for j in self.bases[ii][oo].js:
                        self._slices[(ii, oo)][j] = (
                            out_position,
                            out_position + out_size,
                            in_position,
                            in_position + in_size,
                            basis_count[j],
                            basis_count[j] + self.bases[ii][oo].dim_harmonic(j)
                        )
                        basis_count[j] += self.bases[ii][oo].dim_harmonic(j)
                out_position += out_size
            in_position += in_size

    def dim_harmonic(self, j: Tuple) -> int:
        r'''
            Number of kernel basis elements generated from elements of the steerable filter basis (``self.basis``) which
            transform according to the ``self.basis.group``-irrep identified by ``j``.
        '''
        return self._dim_harmonics[j]

    def compute_harmonics(self, points: torch.Tensor) -> Dict[Tuple, torch.Tensor]:
        r"""
        Pre-compute the sampled steerable filter basis over a set of point.
        This is an alias for ``self.basis.sample_as_dict(points)``.

        .. seealso ::
            :meth:`escnn.kernels.SteerableFiltersBasis.sample_as_dict`.

        """
        return self.basis.sample_as_dict(points)

    def sample(self, points: torch.Tensor, out: torch.Tensor = None) -> torch.Tensor:
        r"""

        Sample the continuous basis elements on the discrete set of points in ``points``.
        Optionally, store the resulting multidimentional array in ``out``.
        
        ``points`` must be an array of shape `(N, d)`, where `N` is the number of points.

        Args:
            points (~torch.Tensor): points where to evaluate the basis elements
            out (~torch.Tensor, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis
            
        """
        assert len(points.shape) == 2
        S = points.shape[0]

        if out is None:
            out = torch.zeros((S, self.dim, self.shape[0], self.shape[1]), device=points.device, dtype=points.dtype)
        else:
            out[:] = 0.
            
        assert out.shape == (S, self.dim, self.shape[0], self.shape[1])

        outs = {}
        B = 0
        for j in self.js:
            outs[j] = out[:, B:B + self.dim_harmonic(j), ...]
            B += self.dim_harmonic(j)

        steerable_basis = self.compute_harmonics(points)
        self.sample_harmonics(steerable_basis, outs)

        return out

    def sample_harmonics(self, points: Dict[Tuple, torch.Tensor], out: Dict[Tuple, torch.Tensor] = None) -> Dict[Tuple, torch.Tensor]:
        r"""
        Sample the continuous basis elements on the discrete set of points.
        Rather than using the points' coordinates, the method directly takes in input the steerable basis elements
        sampled on this points using the method :meth:`escnn.kernels.SteerableKernelBasis.compute_harmonics`.

        Similarly, rather than returning a single tensor containing all sampled basis elements, it groups basis elements
        by the ``G``-irrep acting on them.
        The method returns a dictionary mapping each irrep's ``id`` to a tensor of shape `(N, m, o, i)`, where
        `N` is the number of points,
        `m` is the multiplicity of the irrep (see :meth:`~escnn.kernels.SteerableKernelBasis.dim_harmonic`)
        and `o, i` is the number of input and output channels (see the ``shape`` attribute).

        Optionally, store the resulting tensors in ``out``, rather than allocating new memory.

        Args:
            points (~torch.Tensor): points where to evaluate the basis elements
            out (~torch.Tensor, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """
        if out is None:
            out = {
                j: torch.zeros(
                    (points[j].shape[0], self.dim_harmonic(j), self.shape[0], self.shape[1]),
                    device=points[j].device, dtype=points[j].dtype
                )
                for j in self.js
            }

        for j in self.js:
            if j in out:
                assert out[j].shape == (points[j].shape[0], self.dim_harmonic(j), self.shape[0], self.shape[1])

        if self.A_inv is None and self.B is None:
            out = self._sample_direct_sum(points, out=out)
        else:
            samples = self._sample_direct_sum(points)
            out = self._change_of_basis(samples, out=out)
    
        return out

    def _sample_direct_sum(self, points: Dict[Tuple, torch.Tensor], out: Dict[Tuple, torch.Tensor] = None) -> Dict[Tuple, torch.Tensor]:
    
        if out is None:
            out = {
                j: torch.zeros(
                    (points[j].shape[0], self.dim_harmonic(j), self.shape[0], self.shape[1]),
                    device=points[j].device, dtype=points[j].dtype
                )
                for j in self.js
            }
        # else:
        #     for j in self.js:
        #         if j in out:
        #             out[j][:] = 0.
    
        for j in self.js:
            if j in out:
                assert out[j].shape == (points[j].shape[0], self.dim_harmonic(j), self.shape[0], self.shape[1])

        for ii, in_size in enumerate(self.in_sizes):
            for oo, out_size in enumerate(self.out_sizes):
                if self.bases[ii][oo] is not None:
                    slices = self._slices[(ii, oo)]
                    
                    blocks = {
                        j: out[j][:, b_s:b_e, o_s:o_e, i_s:i_e]
                        for j, (o_s, o_e, i_s, i_e, b_s, b_e) in slices.items()
                    }
                    self.bases[ii][oo].sample_harmonics(points, out=blocks)

        return out
    
    def _change_of_basis(self, samples: Dict[Tuple, torch.Tensor], out: Dict[Tuple, torch.Tensor] = None) -> Dict[Tuple, torch.Tensor]:
        # multiply by the change of basis matrices to transform the irreps basis in the full representations basis

        if out is None:
            out = {j: None for j in self.js}
            
        for j in samples.keys():
            if self.A_inv is not None and self.B is not None:
                out[j][:] = torch.einsum("no,pboi,ij->pbnj", self.B.to(samples[j].dtype), samples[j], self.A_inv.to(samples[j].dtype))
            elif self.A_inv is not None:
                out[j][:] = torch.einsum("pboi,ij->pboj", samples[j], self.A_inv.to(samples[j].dtype))
            elif self.B is not None:
                out[j][:] = torch.einsum("no,pboi->pbni", self.B.to(samples[j].dtype), samples[j])
            else:
                out[j][...] = samples[j]
        
        return out

    def __getitem__(self, idx):
        assert idx < self.dim
        
        j_idx = idx
        for j in self.js:
            dim = self.dim_harmonic(j)
            if j_idx >= dim:
                j_idx -= dim
            else:
                break
        
        assert j_idx < self.dim_harmonic(j), (j_idx, self.dim_harmonic(j))

        count = 0
        for ii in range(len(self.in_sizes)):
            for oo in range(len(self.out_sizes)):
                if self.bases[ii][oo] is not None:
                    dim = self.bases[ii][oo].dim_harmonic(j)

                    rel_idx = j_idx - count
                    if rel_idx >= 0 and rel_idx < dim:
                        
                        attr = dict(self.bases[ii][oo].attrs_j(j, rel_idx))
                        
                        attr["shape"] = self.bases[ii][oo].shape

                        attr["irreps_basis_idx"] = attr["idx"]
                        attr["idx"] = idx
                        attr["j"] = j
                        attr["j_idx"] = j_idx

                        attr["in_irrep"] = self.in_repr.irreps[ii]
                        attr["out_irrep"] = self.out_repr.irreps[oo]
                        
                        attr["in_irrep_idx"] = ii
                        attr["out_irrep_idx"] = oo
                        
                        return attr
                
                    count += dim

    def __iter__(self):
        idx = 0
        for j in self.js:
            j_idx = 0
            for ii in range(len(self.in_sizes)):
                for oo in range(len(self.out_sizes)):
                    basis = self.bases[ii][oo]
                    if basis is not None:
                        
                        for attr in basis.attrs_j_iter(j):
    
                            attr["shape"] = basis.shape
                            
                            attr["in_irrep"] = self.in_repr.irreps[ii]
                            attr["out_irrep"] = self.out_repr.irreps[oo]
                        
                            attr["in_irrep_idx"] = ii
                            attr["out_irrep_idx"] = oo

                            attr["j"] = j
                            attr["j_idx"] = j_idx
                            
                            attr["irreps_basis_idx"] = attr["idx"]
                            attr["idx"] = idx
                            
                            assert idx < self.dim

                            yield attr
                            
                            idx += 1
                            j_idx += 1

    def __eq__(self, other):
        if not isinstance(other, SteerableKernelBasis):
            return False
        elif self.basis != other.basis or self.in_repr != other.in_repr or self.out_repr != other.out_repr:
            return False
        else:
            sbk1 = sorted(self.irreps_bases.keys())
            sbk2 = sorted(other.irreps_bases.keys())
            if sbk1 != sbk2:
                return False
            
            for irreps, basis in self.irreps_bases.items():
                if basis != other.irreps_bases[irreps]:
                    return False
            
            return True

    def __hash__(self):
        h = hash(self.in_repr) + hash(self.out_repr) + hash(self.basis)
        for basis in self.irreps_bases.items():
            h += hash(basis)
        return h
        
        






import torch
import math
import numpy as np

from torch.optim import Adam, SGD


def _normalize(X: torch.Tensor) -> torch.Tensor:
    X /= torch.norm(X, dim=1, keepdim=True)


def _distance_sphere(X: torch.Tensor, Y: torch.Tensor = None, type: str = 'euclidean') -> torch.Tensor:

    if Y is None:
        Y = X

    with torch.no_grad():
        assert torch.isclose(torch.norm(X, dim=1), torch.tensor(1., dtype=X.dtype)).all()
        assert torch.isclose(torch.norm(Y, dim=1), torch.tensor(1., dtype=Y.dtype)).all()

    if type == 'cos':

        cos = X @ Y.T

        dist = torch.zeros_like(cos)

        ATOL = 1e-15
        grad_mask_0 = torch.isclose(cos, torch.tensor(1., dtype=cos.dtype), atol=ATOL)
        grad_mask_1 = torch.isclose(cos, torch.tensor(-1., dtype=cos.dtype), atol=ATOL)
        grad_mask = grad_mask_0 | grad_mask_1

        dist[~grad_mask] = torch.acos(cos[~grad_mask]) / math.pi
        dist[grad_mask_0] = 0.
        dist[grad_mask_1] = 1.

        return dist

    elif type == 'euclidean':
        dist = torch.norm(X.unsqueeze(1) - Y, dim=2, keepdim=False)
        assert dist.shape == (X.shape[0], Y.shape[0])
        return dist
    else:
        raise ValueError()


def _potential_sphere(X: torch.Tensor, distance: str = 'euclidean') -> torch.Tensor:

    dist = _distance_sphere(X, type=distance)

    mask = torch.triu(torch.ones(X.shape[0], X.shape[0], dtype=torch.bool), diagonal=1)

    dist = dist[mask]

    potential = 1. / dist
    return potential.sum()


def _thomson_sphere(N: int, lr: float = 1e-1, optimizer: str = 'Adam', rng: np.random.RandomState = None, verbose: int = 0) -> torch.Tensor:

    if rng is None:
        rng = np.random

    if torch.cuda.is_available():
        device = 'cuda'
    else:
        device = 'cpu'

    X = torch.tensor(
        rng.randn(N, 3).astype(np.float32),
        requires_grad=True, device=device, dtype=torch.float32
    )

    if optimizer == 'Adam':
        optim = Adam(
            params=(X,),
            lr=lr,
            weight_decay=0.
        )
    elif optimizer == 'SGD':
        optim = SGD(
            params=(X,),
            lr=lr,
            momentum=.2,
            weight_decay=0.
        )
    else:
        raise ValueError()

    last_U = torch.tensor(N**2, dtype=X.dtype, device=device)
    best_U = last_U
    best_X = None

    for i in range(5000):

        with torch.no_grad():
            _normalize(X)

        optim.zero_grad()

        U = _potential_sphere(X, 'euclidean')

        if verbose > 1:
            print(f'{i}: {U.item()}')

        if torch.isclose(U, last_U, atol=1e-16, rtol=0.):
            break

        last_U = U.detach()

        if last_U < best_U:
            best_U = last_U
            best_X = X.detach().cpu()

        U.backward()
        optim.step()

    X = best_X
    _normalize(X)

    if verbose > 0:
        # Uc = _potential_sphere(X, 'cos')
        Ue = _potential_sphere(X, 'euclidean')

        # print(f'Geodesic Distance: {Uc.item()}')
        print(f'Euclidean Distance: {Ue.item()}')

    return X.detach().cpu()


def _distance_so3(X: torch.Tensor, Y: torch.Tensor = None, type: str = 'euclidean') -> torch.Tensor:

    if Y is None:
        Y = X

    d1 = _distance_sphere(X, Y, type=type)
    d2 = _distance_sphere(X, -Y, type=type)

    dist = torch.min(d1, d2)

    return dist


def _potential_so3(X: torch.Tensor, distance: str = 'euclidean') -> torch.Tensor:

    dist = _distance_so3(X, type=distance)

    mask = torch.triu(torch.ones(X.shape[0], X.shape[0], dtype=torch.bool), diagonal=1)

    dist = dist[mask]

    potential = 1. / dist
    return potential.sum()


def _thomson_so3(N: int, lr: float = 1e-1, optimizer: str = 'Adam', rng: np.random.RandomState = None, verbose: int = 0) -> torch.Tensor:

    if rng is None:
        rng = np.random

    if torch.cuda.is_available():
        device = 'cuda'
    else:
        device = 'cpu'

    X = torch.tensor(
        rng.randn(N, 4).astype(np.float32),
        requires_grad=True, device=device, dtype=torch.float32
    )

    if optimizer == 'Adam':
        optim = Adam(
            params=(X,),
            lr=lr,
            weight_decay=0.
        )
    elif optimizer == 'SGD':
        optim = SGD(
            params=(X,),
            lr=lr,
            momentum=.2,
            weight_decay=0.
        )
    else:
        raise ValueError()

    last_U = torch.tensor(N**2, dtype=X.dtype, device=device)
    best_U = last_U
    best_X = None

    for i in range(5000):

        with torch.no_grad():
            _normalize(X)

        optim.zero_grad()

        U = _potential_so3(X, 'euclidean')

        if verbose > 1:
            print(f'{i}: {U.item()}')

        if torch.isclose(U, last_U, atol=1e-16, rtol=0.):
            break

        last_U = U.detach()

        if last_U < best_U:
            best_U = last_U
            best_X = X.detach().cpu()

        U.backward()
        optim.step()

    X = best_X
    _normalize(X)

    if verbose > 0:

        # Uc = _potential_so3(X, 'cos')
        Ue = _potential_so3(X, 'euclidean')

        # print(f'Geodesic Distance: {Uc.item()}')
        print(f'Euclidean Distance: {Ue.item()}')

    return X.detach().cpu()


from scipy.spatial.transform import Rotation


def _thomson_cube_sphere(N: int, lr: float = 1e-1, optimizer: str = 'Adam', rng: np.random.RandomState = None, verbose: int = 0) -> torch.Tensor:

    if rng is None:
        rng = np.random

    if torch.cuda.is_available():
        device = 'cuda'
    else:
        device = 'cpu'

    X = torch.tensor(
        rng.randn(N, 3).astype(np.float32),
        requires_grad=True, device=device, dtype=torch.float32
    )
    cube = torch.tensor(
        Rotation.create_group('O').as_matrix(),
        device=X.device,
        dtype=X.dtype
    )
    assert cube.shape == (24, 3, 3)

    def _cubify(_X: torch.Tensor) -> torch.Tensor:
        return torch.einsum('gij,sj->gsi', cube.to(_X.device), _X).reshape(-1, 3)

    if optimizer == 'Adam':
        optim = Adam(
            params=(X,),
            lr=lr,
            weight_decay=0.
        )
    elif optimizer == 'SGD':
        optim = SGD(
            params=(X,),
            lr=lr,
            momentum=.2,
            weight_decay=0.
        )
    else:
        raise ValueError()

    last_U = torch.tensor((24*N)**2, dtype=X.dtype, device=device)
    best_U = last_U
    best_X = None

    for i in range(5000):

        with torch.no_grad():
            _normalize(X)

        optim.zero_grad()

        U = _potential_sphere(_cubify(X), 'euclidean')

        if verbose > 1:
            print(f'{i}: {U.item()}')

        if torch.isclose(U, last_U, atol=1e-16, rtol=0.):
            break

        last_U = U.detach()

        if last_U < best_U:
            best_U = last_U
            best_X = X.detach().cpu()

        U.backward()
        optim.step()

    X = best_X
    _normalize(X)

    if verbose > 0:
        # Uc = _potential_sphere(_cubify(X), 'cos')
        Ue = _potential_sphere(_cubify(X), 'euclidean')

        # print(f'Geodesic Distance: {Uc.item()}')
        print(f'Euclidean Distance: {Ue.item()}')

    return _cubify(X).detach().cpu()


def _thomson_cube_so3(N: int, lr: float = 1e-1, optimizer: str = 'Adam', rng: np.random.RandomState = None, verbose: int = 0) -> torch.Tensor:

    if rng is None:
        rng = np.random

    if torch.cuda.is_available():
        device = 'cuda'
    else:
        device = 'cpu'

    X = torch.tensor(
        rng.randn(N, 4).astype(np.float32),
        requires_grad=True, device=device, dtype=torch.float32
    )

    cube = torch.tensor(
        Rotation.create_group('O').as_quat(),
        device=X.device,
        dtype=X.dtype
    )
    assert cube.shape == (24, 4)

    quaternion_basis = torch.stack([
        torch.tensor([
            [0, 0, 0, 1],
            [0, 0, -1, 0],
            [0, 1, 0, 0],
            [-1, 0, 0, 0],
        ], dtype=X.dtype),
        torch.tensor([
            [0, 0, 1, 0],
            [0, 0, 0, 1],
            [-1, 0, 0, 0],
            [0, -1, 0, 0],
        ], dtype=X.dtype),
        torch.tensor([
            [0, -1, 0, 0],
            [1, 0, 0, 0],
            [0, 0, 0, 1],
            [0, 0, -1, 0],
        ], dtype=X.dtype),
        torch.eye(4, dtype=X.dtype),
    ], dim=0).to(device=X.device)

    def _cubify(_X: torch.Tensor) -> torch.Tensor:
        return torch.einsum('bij,gb,sj->gsi', quaternion_basis.to(_X.device), cube.to(_X.device), _X).reshape(-1, 4)

    if optimizer == 'Adam':
        optim = Adam(
            params=(X,),
            lr=lr,
            weight_decay=0.
        )
    elif optimizer == 'SGD':
        optim = SGD(
            params=(X,),
            lr=lr,
            momentum=.2,
            weight_decay=0.
        )
    else:
        raise ValueError()

    last_U = torch.tensor((24*N)**2, dtype=X.dtype, device=device)
    best_U = last_U
    best_X = None

    for i in range(5000):

        with torch.no_grad():
            _normalize(X)

        optim.zero_grad()

        U = _potential_so3(_cubify(X), 'euclidean')

        if verbose > 1:
            print(f'{i}: {U.item()}')

        if torch.isclose(U, last_U, atol=1e-16, rtol=0.):
            break

        last_U = U.detach()

        if last_U < best_U:
            best_U = last_U
            best_X = X.detach().cpu()

        U.backward()
        optim.step()

    X = best_X
    _normalize(X)

    if verbose > 0:

        # Uc = _potential_so3(_cubify(X), 'cos')
        Ue = _potential_so3(_cubify(X), 'euclidean')

        # print(f'Geodesic Distance: {Uc.item()}')
        print(f'Euclidean Distance: {Ue.item()}')

    return _cubify(X).detach().cpu()

########################################################################################################################

from joblib import Memory
# import os
# cache_sphere = Memory(os.path.join(os.path.dirname(__file__), '_jl_thomson_sphere'), verbose=2)
# cache_so3 = Memory(os.path.join(os.path.dirname(__file__), '_jl_thomson_so3'), verbose=2)
# cache_cube_sphere = Memory(os.path.join(os.path.dirname(__file__), '_jl_thomson_cube_sphere'), verbose=2)
# cache_cube_so3 = Memory(os.path.join(os.path.dirname(__file__), '_jl_thomson_cube_so3'), verbose=2)

from escnn.group import __cache_path__
cache = Memory(__cache_path__, verbose=0)


# @cache_sphere.cache
@cache.cache
def thomson_sphere(N: int) -> np.ndarray:
    attempts = 20
    verbose = 0

    # to ensure a determinist behaviour of this method
    rng = np.random.RandomState(42)

    best_U = N ** 2
    best_X = None
    for i in range(attempts):
        X = _thomson_sphere(N, lr=1e-1, optimizer='Adam', rng=rng, verbose=verbose - 1)
        U = _potential_sphere(X).item()

        if U < best_U:
            best_U = U
            best_X = X

    if verbose > 0:
        print(f'Best Potential: {best_U}')

    return best_X.numpy()


# @cache_so3.cache
@cache.cache
def thomson_so3(N: int) -> np.ndarray:

    attempts = 10
    verbose = 0

    # to ensure a determinist behaviour of this method
    rng = np.random.RandomState(42)

    best_U = N**2
    best_X = None
    for i in range(attempts):
        X = _thomson_so3(N, lr=5e-2, optimizer='Adam', rng=rng, verbose=verbose - 1)
        U = _potential_so3(X).item()

        if U < best_U:
            best_U = U
            best_X = X

    if verbose > 0:
        print(f'Best Potential: {best_U}')

    return best_X.numpy()


# @cache_cube_sphere.cache
@cache.cache
def thomson_cube_sphere(N: int) -> np.ndarray:
    attempts = 20
    verbose = 0

    # to ensure a determinist behaviour of this method
    rng = np.random.RandomState(42)

    best_U = (24*N) ** 2
    best_X = None
    for i in range(attempts):
        X = _thomson_cube_sphere(N, lr=1e-1, optimizer='Adam', rng=rng, verbose=verbose - 1)
        U = _potential_sphere(X).item()

        if U < best_U:
            best_U = U
            best_X = X

    if verbose > 0:
        print(f'Best Potential: {best_U}')

    return best_X.numpy()


# @cache_cube_so3.cache
@cache.cache
def thomson_cube_so3(N: int) -> np.ndarray:

    attempts = 10
    verbose = 0

    # to ensure a determinist behaviour of this method
    rng = np.random.RandomState(42)

    best_U = (24*N)**2
    best_X = None
    for i in range(attempts):
        X = _thomson_cube_so3(N, lr=5e-2, optimizer='Adam', rng=rng, verbose=verbose - 1)
        U = _potential_so3(X).item()

        if U < best_U:
            best_U = U
            best_X = X

    if verbose > 0:
        print(f'Best Potential: {best_U}')

    return best_X.numpy()


if __name__ == '__main__':

    # X = thomson_sphere(120)
    # X = thomson_so3(130)
    # X = _thomson_cube_so3(3, verbose=3)
    X = _thomson_cube_sphere(1, verbose=3)
    # X = _thomson_so3(72, verbose=3)




from abc import ABC, abstractmethod

from escnn.nn import FieldType
from escnn.nn import GeometricTensor
from escnn.group import Representation
from escnn.kernels import KernelBasis, EmptyBasisException
from escnn.gspaces import *

from ..equivariant_module import EquivariantModule

from escnn.nn.modules.basismanager import BlocksBasisSampler

from typing import Callable, Tuple, Dict, Union

import torch

import torch_geometric

from torch.nn import Parameter
import numpy as np
import math


__all__ = ["_RdPointConv"]


class _RdPointConv(torch_geometric.nn.MessagePassing, EquivariantModule, ABC):
    
    def __init__(self,
                 in_type: FieldType,
                 out_type: FieldType,
                 d: int,
                 groups: int = 1,
                 bias: bool = True,
                 basis_filter: Callable[[dict], bool] = None,
                 recompute: bool = False,
                 ):

        r"""

        Abstract class which implements a general G-steerable convolution, mapping between the input and output
        :class:`~escnn.nn.FieldType` s specified by the parameters ``in_type`` and ``out_type``.
        This operation is equivariant under the action of :math:`\R^d\rtimes G` where :math:`G` is the
        :attr:`escnn.nn.FieldType.fibergroup` of ``in_type`` and ``out_type``.

        This class implements convolution with steerable filters over sparse planar geometric graphs.
        Instead, :class:`~escnn.nn.modules.conv._RdConv` implements an equivalent convolution layer over a pixel/voxel
        grid. See the documentation of :class:`~escnn.nn.modules.conv._RdConv` for more details about equivariance and
        steerable convolution.

        The input of this module is a geometric graph, i.e. a graph whose nodes are associated with ``d``-dimensional
        coordinates in :math:`\R^d`.
        The nodes' coordinates should be stored in the ``coords`` attribute of the input
        :class:`~escnn.nn.GeometricTensor`.
        The adjacency of the graph should be passed as a second input tensor ``edge_index``, like commonly done in
        :class:`~torch_geometric.nn.conv.message_passing.MessagePassing`.
        See :meth:`~escnn.nn.modules.pointconv._RdPointConv.forward`.

        In each forward pass, the module computes the relative coordinates of the points on the edges and samples each
        filter in the basis of G-steerable kernels at these relative locations.
        The basis filters are expanded using the learnable weights and used to perform convolution over the graph in the
        message passing framework.
        Optionally, the relative coordinates can be pre-computed and passed in the input ``edge_delta`` tensor.

        .. note ::
            In practice, we first apply the basis filters on the input features and then combine the responses via the
            learnable weights. See also :meth:`~escnn.nn.modules.basismanager.BlocksBasisSampler.compute_messages`.

        .. warning ::

            When :meth:`~torch.nn.Module.eval()` is called, the bias is built with the current trained weights and stored
            for future reuse such that no overhead of expanding the bias remains.

            When :meth:`~torch.nn.Module.train()` is called, the attribute
            :attr:`~escnn.nn.modules.pointconv._RdPointConv.expanded_bias` is discarded to avoid situations of mismatch
            with the learnable expansion coefficients. See also :meth:`escnn.nn.modules.pointconv._RdPointConv.train`.

            This behaviour can cause problems when storing the :meth:`~torch.nn.Module.state_dict` of a model while in
            a mode and lately loading it in a model with a different mode, as the attributes of the class change.
            To avoid this issue, we recommend converting the model to eval mode before storing or loading the state
            dictionary.

        .. warning ::
            We don't support ``groups > 1`` yet.
            We include this parameter for future compatibility.

        Args:
            in_type (FieldType): the type of the input field, specifying its transformation law
            out_type (FieldType): the type of the output field, specifying its transformation law
            d (int): dimensionality of the base space (2 for images, 3 for volumes)
            groups (int, optional): number of blocked connections from input channels to output channels.
                                    It allows depthwise convolution. When used, the input and output types need to be
                                    divisible in ``groups`` groups, all equal to each other.
                                    Default: ``1``.
            bias (bool, optional): Whether to add a bias to the output (only to fields which contain a
                    trivial irrep) or not. Default ``True``
            basis_filter (callable, optional): function which takes as input a descriptor of a basis element
                    (as a dictionary) and returns a boolean value: whether to preserve (``True``) or discard (``False``)
                    the basis element. By default (``None``), no filtering is applied.
            recompute (bool, optional): if ``True``, recomputes a new basis for the equivariant kernels.
                    By Default (``False``), it  caches the basis built or reuse a cached one, if it is found.

        Attributes:

            ~.weights (torch.Tensor): the learnable parameters which are used to expand the kernel
            ~.bias (torch.Tensor): the learnable parameters which are used to expand the bias, if ``bias=True``
            ~.expanded_bias (torch.Tensor): the equivariant bias which is summed to the output, obtained by expanding
                                    the parameters in :attr:`~escnn.nn.modules.pointconv._RdPointConv.bias`

        """

        assert in_type.gspace == out_type.gspace
        assert isinstance(in_type.gspace, GSpace)
        assert d >= in_type.gspace.dimensionality

        super(_RdPointConv, self).__init__(aggr='mean')

        self.d = d
        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = out_type

        assert self.space.dimensionality == self.d
        
        self.groups = groups

        if groups > 1:
            # Check the input and output classes can be split in `groups` groups, all equal to each other
            # first, check that the number of fields is divisible by `groups`
            assert len(in_type) % groups == 0
            assert len(out_type) % groups == 0
            in_size = len(in_type) // groups
            out_size = len(out_type) // groups
            
            # then, check that all groups are equal to each other, i.e. have the same types in the same order
            assert all(in_type.representations[i] == in_type.representations[i % in_size] for i in range(len(in_type)))
            assert all(out_type.representations[i] == out_type.representations[i % out_size] for i in range(len(out_type)))
            
            # finally, retrieve the type associated to a single group in input.
            # this type will be used to build a smaller kernel basis and a smaller filter
            # as in PyTorch, to build a filter for grouped convolution, we build a filter which maps from one input
            # group to all output groups. Then, PyTorch's standard convolution routine interpret this filter as `groups`
            # different filters, each mapping an input group to an output group.
            in_type = in_type.index_select(list(range(in_size)))
        
        if bias:
            # bias can be applied only to trivial irreps inside the representation
            # to apply bias to a field we learn a bias for each trivial irreps it contains
            # and, then, we transform it with the change of basis matrix to be able to apply it to the whole field
            # this is equivalent to transform the field to its irreps through the inverse change of basis,
            # sum the bias only to the trivial irrep and then map it back with the change of basis
            
            # count the number of trivial irreps
            trivials = 0
            for r in self.out_type:
                for irr in r.irreps:
                    if self.out_type.fibergroup.irrep(*irr).is_trivial():
                        trivials += 1
            
            # if there is at least 1 trivial irrep
            if trivials > 0:
                
                # matrix containing the columns of the change of basis which map from the trivial irreps to the
                # field representations. This matrix allows us to map the bias defined only over the trivial irreps
                # to a bias for the whole field more efficiently
                bias_expansion = torch.zeros(self.out_type.size, trivials)
                
                p, c = 0, 0
                for r in self.out_type:
                    pi = 0
                    for irr in r.irreps:
                        irr = self.out_type.fibergroup.irrep(*irr)
                        if irr.is_trivial():
                            bias_expansion[p:p+r.size, c] = torch.tensor(r.change_of_basis[:, pi])
                            c += 1
                        pi += irr.size
                    p += r.size
                
                self.register_buffer("bias_expansion", bias_expansion)
                self.bias = Parameter(torch.zeros(trivials), requires_grad=True)
                self.register_buffer("expanded_bias", torch.zeros(out_type.size))
            else:
                self.bias = None
                self.expanded_bias = None
        else:
            self.bias = None
            self.expanded_bias = None

        # TODO support `groups` arg for conv
        if groups != 1:
            raise NotImplementedError(f'`groups !=1` not supported yet!')

        # BlocksBasisSampler: submodule which takes care of building the filter
        self._basissampler = BlocksBasisSampler(in_type.representations, out_type.representations,
                                                    self._build_kernel_basis,
                                                    basis_filter=basis_filter,
                                                    recompute=recompute)

        if self.basissampler.dimension() == 0:
            raise ValueError('''
                The basis for the steerable filter is empty!
                Tune the `frequencies_cutoff`, `kernel_size`, `rings` or `basis_filter` parameters to allow
                for a larger basis.
            ''')

        self.weights = Parameter(torch.zeros(self.basissampler.dimension()), requires_grad=True)

    @abstractmethod
    def _build_kernel_basis(self, in_repr: Representation, out_repr: Representation) -> KernelBasis:
        raise NotImplementedError

    @property
    def basissampler(self) -> BlocksBasisSampler:
        r"""

        Submodule which takes care of sampling the steerable filters.

        It is used to sample the G-steerable basis on the relative coordinates along the edges of a geometric graph and,
        then, expand the kernel in the sampled basis using the learned ``weights``.
        See also :meth:`~escnn.nn.modules.basismanager.BlockBasisSampler.forward`.

        In practice, this submodule is also used to directly compute the messages via
        :meth:`~escnn.nn.modules.basismanager.BlockBasisSampler.compute_messages`: first, the basis filters are applied
        on the input features and, then, the responses are combined using the learnable weights.

        """
        return self._basissampler

    def expand_bias(self) -> torch.Tensor:
        r"""

        Expand the bias in terms of :class:`~escnn.nn._RdPointConv.bias`.

        Returns:
            the expanded bias

        """
        if self.bias is None:
            _bias = None
        else:
            _bias = self.bias_expansion @ self.bias

        return _bias

    def expand_filter(self, points: Union[torch.Tensor, Dict[Tuple, torch.Tensor]]) -> torch.Tensor:
        r"""

        Expand the filter in terms of :class:`~escnn.nn._RdPointConv.weights`.

        Returns:
            the expanded filter sampled on the input points

        """

        return self.basissampler(self.weights, points)

    def expand_parameters(self, points: Union[torch.Tensor, Dict[Tuple, torch.Tensor]]) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""

        Expand the filter in terms of the :attr:`~escnn.nn._RdConv.weights` and the
        expanded bias in terms of :class:`~escnn.nn._RdConv.bias`.

        Returns:
            the expanded filter and bias

        """

        _filter = self.expand_filter(points)
        _bias = self.expand_bias()

        return _filter, _bias

    def forward(self, x: GeometricTensor, edge_index: torch.Tensor, edge_delta: torch.Tensor = None):
        r"""
        Convolve the input with the expanded filter and bias.

        This method is based on PyTorch Geometric's :class:`~torch_geometric.nn.conv.message_passing.MessagePassing`,
        i.e. it uses :meth:`~torch_geometric.nn.conv.message_passing.MessagePassing.propagate` to send the messages
        computed in :meth:`~escnn.nn.modules.pointconv._RdPointConv.message`.

        The input tensor ``input`` represents a feature field over the nodes of a geometric graph.
        Hence, the ``coords`` attribute of ``input`` should contain the ``d``-dimensional coordinates of each node (see
        :class:`~escnn.nn.GeometricTensor`).

        The tensor ``edge_index`` must be a :class:`torch.LongTensor` of shape ``(2, m)``, representing ``m`` edges.

        Mini-batches containing multiple graphs can be constructed as in
        `Pytorch Geometric <https://pytorch-geometric.readthedocs.io/en/latest/notes/batching.html>`_ by merging the
        graphs in a unique, disconnected, graph.

        Args:
            input (GeometricTensor): input feature field transforming according to ``in_type``.
            edge_index (torch.Tensor): tensor representing the connectivity of the graph.
            edge_delta (torch.Tensor, optional): the relative coordinates of the nodes on each edge. If not passed, it
                    is automatically computed using ``input.coords`` and ``edge_index``.

        Returns:
            output feature field transforming according to ``out_type``

        """
        assert isinstance(x, GeometricTensor)
        assert x.type == self.in_type

        assert len(edge_index.shape) == 2
        assert edge_index.shape[0] == 2

        if edge_delta is None:
            pos = x.coords
            row, cols = edge_index
            edge_delta = pos[row] - pos[cols]

        out = self.propagate(edge_index, x=x.tensor, edge_delta=edge_delta)

        if not self.training:
            _bias = self.expanded_bias
        else:
            # retrieve the bias
            _bias = self.expand_bias()

        if _bias is not None:
            out += _bias

        out = GeometricTensor(out, self.out_type, coords=x.coords)
        
        return out

    def message(self, x_j: torch.Tensor, edge_delta: torch.Tensor=None) -> torch.Tensor:
        r"""

        This methods computes the message from the input node ``j`` to the output node ``i`` of each edge in
        ``edge_index``.

        The message is equal to the product of the filter evaluated on the relative coordinate along an edge with
        the feature vector on the input node of the edge.

        """
        return self.basissampler.compute_messages(self.weights, x_j, edge_delta, conv_first=True)

    def train(self, mode=True):
        r"""

        If ``mode=True``, the method sets the module in training mode and discards the
        :attr:`~escnn.nn._RdPointConv.expanded_bias` attribute.

        If ``mode=False``, it sets the module in evaluation mode. Moreover, the method builds the bias
        using the current values of the trainable parameters and store it :attr:`~escnn.nn._RdConv.expanded_bias`
        such that it is not recomputed at each forward pass.

        .. warning ::

            This behaviour can cause problems when storing the :meth:`~torch.nn.Module.state_dict` of a model while in
            a mode and lately loading it in a model with a different mode, as the attributes of this class change.
            To avoid this issue, we recommend converting the model to eval mode before storing or loading the state
            dictionary.

        Args:
            mode (bool, optional): whether to set training mode (``True``) or evaluation mode (``False``).
                                   Default: ``True``.

        """

        if mode:
            if hasattr(self, "expanded_bias"):
                del self.expanded_bias
        elif self.training:
            # avoid re-computation of the filter and the bias on multiple consecutive calls of `.eval()`

            _bias = self.expand_bias()

            if _bias is not None:
                self.register_buffer("expanded_bias", _bias)
            else:
                self.expanded_bias = None

        return super(_RdPointConv, self).train(mode)

    def evaluate_output_shape(self, input_shape: Tuple) -> Tuple:
        assert len(input_shape) == 3
        assert input_shape[1] == self.in_type.size

        return input_shape[0], self.out_type.size, input_shape[2]

    def __repr__(self):
        extra_lines = []
        extra_repr = self.extra_repr()
        if extra_repr:
            extra_lines = extra_repr.split('\n')

        main_str = self._get_name() + '('
        if len(extra_lines) == 1:
            main_str += extra_lines[0]
        else:
            main_str += '\n  ' + '\n  '.join(extra_lines) + '\n'

        main_str += ')'
        return main_str

    def extra_repr(self):
        s = ('{in_type}, {out_type}')
        if self.groups != 1:
            s += ', groups={groups}'
        if self.bias is None:
            s += ', bias=False'
        return s.format(**self.__dict__)

    def check_equivariance(self, atol: float = 1e-5, rtol: float = 1e-6, assertion: bool = True, verbose: bool = True):
    
        # np.set_printoptions(precision=5, threshold=30 *self.in_type.size**2, suppress=False, linewidth=30 *self.in_type.size**2)
    
        P = 30
    
        pos = torch.randn(P, self.d)
        x = torch.randn(P, self.in_type.size)
        x = GeometricTensor(x, self.in_type, pos)
    
        distance = torch.norm(pos.unsqueeze(1) - pos, dim=2, keepdim=False)
    
        thr = sorted(distance.view(-1).tolist())[int(P**2//16)]
        edge_index = torch.nonzero(distance < thr).T.contiguous()

        errors = []
    
        for el in self.space.testing_elements:
        
            out1 = self(x, edge_index).transform(el).tensor.detach().numpy()
            out2 = self(x.transform(el), edge_index).tensor.detach().numpy()
        
            errs = np.abs(out1 - out2)
        
            esum = np.maximum(np.abs(out1), np.abs(out2))
            esum[esum == 0.0] = 1
        
            relerr = errs / esum
        
            # if verbose:
            #     print(el)
            #     print(relerr.max(), relerr.mean(), relerr.var(), errs.max(), errs.mean(), errs.var())
        
            tol = rtol * esum + atol
        
            if np.any(errs > tol) and verbose:
                print(out1[errs > tol])
                print(out2[errs > tol])
                print(tol[errs > tol])
        
            if assertion:
                assert np.all(
                    errs < tol), 'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}'.format(
                    el, errs.max(), errs.mean(), errs.var())
        
            errors.append((el, errs.mean()))
    
        return errors






from torch.nn.functional import conv_transpose2d

import escnn.nn
from escnn.nn import FieldType
from escnn.nn import GeometricTensor
from escnn.gspaces import *
from escnn.group import Representation, Group
from escnn.kernels import KernelBasis

from .rd_transposed_convolution import _RdConvTransposed
from .r2convolution import compute_basis_params


from typing import Callable, Union, Tuple, List

import torch
import numpy as np


__all__ = ["R2ConvTransposed"]


class R2ConvTransposed(_RdConvTransposed):
    
    def __init__(self,
                 in_type: FieldType,
                 out_type: FieldType,
                 kernel_size: int,
                 padding: int = 0,
                 output_padding: int = 0,
                 stride: int = 1,
                 dilation: int = 1,
                 groups: int = 1,
                 bias: bool = True,
                 sigma: Union[List[float], float] = None,
                 frequencies_cutoff: Union[float, Callable[[float], int]] = None,
                 rings: List[float] = None,
                 maximum_offset: int = None,
                 recompute: bool = False,
                 basis_filter: Callable[[dict], bool] = None,
                 initialize: bool = True,
                 ):
        r"""
        
        Transposed G-steerable planar convolution layer.

        .. warning ::

            This class implements a *discretized* convolution operator over a discrete grid.
            This means that equivariance to continuous symmetries is *not* perfect.
            In practice, by using sufficiently band-limited filters, the equivariance error introduced by the
            discretization of the filters and the features is contained, but some design choices may have a negative
            effect on the overall equivariance of the architecture.

            We provide some :doc:`practical notes <conv_notes>` on using this discretized
            convolution module.


        .. warning ::
            
            Transposed convolution can produce artifacts which can harm the overall equivariance of the model.
            We suggest using :class:`~escnn.nn.R2Upsampling` combined with :class:`~escnn.nn.R2Conv` to perform
            upsampling.
        
        .. seealso ::
            For additional information about the parameters and the methods of this class, see :class:`escnn.nn.R2Conv`.
            The two modules are essentially the same, except for the type of convolution used.
            
        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.
            
        Args:
            in_type (FieldType): the type of the input field
            out_type (FieldType): the type of the output field
            kernel_size (int): the size of the filter
            padding(int, optional): implicit zero paddings on both sides of the input. Default: ``0``
            output_padding(int, optional): implicit zero paddings on both sides of the input. Default: ``0``
            stride(int, optional): the stride of the convolving kernel. Default: ``1``
            dilation(int, optional): the spacing between kernel elements. Default: ``1``
            groups (int, optional): number of blocked connections from input channels to output channels.
                                    Default: ``1``.
            bias (bool, optional): Whether to add a bias to the output (only to fields which contain a
                    trivial irrep) or not. Default ``True``
            initialize (bool, optional): initialize the weights of the model. Default: ``True``
            
        """
        assert isinstance(in_type.gspace, GSpace2D)
        assert isinstance(out_type.gspace, GSpace2D)

        basis_filter, self._rings, self._sigma, self._maximum_frequency = compute_basis_params(
            kernel_size, frequencies_cutoff, rings, sigma, dilation, basis_filter
        )

        super(R2ConvTransposed, self).__init__(
            in_type,
            out_type,
            2,
            kernel_size,
            padding,
            output_padding,
            stride,
            dilation,
            groups,
            bias,
            basis_filter,
            recompute,
        )

        if initialize:
            # by default, the weights are initialized with a generalized form of He's weight initialization
            escnn.nn.init.generalized_he_init(self.weights.data, self.basisexpansion)

    def _build_kernel_basis(self, in_repr: Representation, out_repr: Representation) -> KernelBasis:
        return self.space.build_kernel_basis(in_repr, out_repr, self._sigma, self._rings,
                                             maximum_frequency=self._maximum_frequency
                                             )

    def forward(self, input: GeometricTensor):
        assert input.type == self.in_type

        if not self.training:
            _filter = self.filter
            _bias = self.expanded_bias
        else:
            # retrieve the filter and the bias
            _filter, _bias = self.expand_parameters()
        
        # use it for convolution and return the result
        output = conv_transpose2d(
                        input.tensor, _filter,
                        padding=self.padding,
                        output_padding=self.output_padding,
                        stride=self.stride,
                        dilation=self.dilation,
                        groups=self.groups,
                        bias=_bias)
        
        return GeometricTensor(output, self.out_type, coords=None)
    
    def check_equivariance(self, atol: float = 0.1, rtol: float = 0.1, assertion: bool = True, verbose: bool = True):

        _filter, _bias = self.expand_parameters()
        print(self.in_type.size, self.out_type.size, _filter.shape)
        
        # np.set_printoptions(precision=5, threshold=30 *self.in_type.size**2, suppress=False, linewidth=30 *self.in_type.size**2)

        feature_map_size = 33
        last_downsampling = 5
        first_downsampling = 5
        
        initial_size = (feature_map_size * last_downsampling - 1 + self.kernel_size) * first_downsampling
        
        c = self.in_type.size
    
        # x = torch.randn(3, c, 10, 10)
        
        import matplotlib.image as mpimg
        from skimage.measure import block_reduce
        from skimage.transform import resize

        # x = mpimg.imread('../group/testimage.jpeg').transpose((2, 0, 1))[np.newaxis, 0:c, :, :]
        import scipy
        x = scipy.datasets.face().transpose((2, 0, 1))[np.newaxis, 0:c, :, :]

        x = resize(
            x,
            (x.shape[0], x.shape[1], initial_size, initial_size),
            anti_aliasing=True
        )
        
        x = x / 255.0 - 0.5
        
        if x.shape[1] < c:
            to_stack = [x for i in range(c // x.shape[1])]
            if c % x.shape[1] > 0:
                to_stack += [x[:, :(c % x.shape[1]), ...]]
    
            x = np.concatenate(to_stack, axis=1)
    
        x = GeometricTensor(torch.FloatTensor(x), self.in_type)
        
        def shrink(t: GeometricTensor, s) -> GeometricTensor:
            return GeometricTensor(torch.FloatTensor(block_reduce(t.tensor.detach().numpy(), s, func=np.mean)), t.type)
        
        errors = []
    
        for el in self.space.testing_elements:
            
            out1 = self(shrink(x, (1, 1, 5, 5))).transform(el).tensor.detach().numpy()
            out2 = self(shrink(x.transform(el), (1, 1, 5, 5))).tensor.detach().numpy()
            
            out1 = block_reduce(out1, (1, 1, 5, 5), func=np.mean)
            out2 = block_reduce(out2, (1, 1, 5, 5), func=np.mean)
            
            b, c, h, w = out2.shape

            center_mask = np.zeros((2, h, w))
            center_mask[1, :, :] = np.arange(0, w) - w / 2
            center_mask[0, :, :] = np.arange(0, h) - h / 2
            center_mask[0, :, :] = center_mask[0, :, :].T
            center_mask = center_mask[0, :, :] ** 2 + center_mask[1, :, :] ** 2 < (h / 4) ** 2

            out1 = out1[..., center_mask]
            out2 = out2[..., center_mask]
            
            out1 = out1.reshape(-1)
            out2 = out2.reshape(-1)
            
            errs = np.abs(out1 - out2)

            esum = np.maximum(np.abs(out1), np.abs(out2))
            esum[esum == 0.0] = 1

            relerr = errs / esum
    
            if verbose:
                print(el, relerr.max(), relerr.mean(), relerr.var(), errs.max(), errs.mean(), errs.var())
        
            # tol = rtol*(np.abs(out1) + np.abs(out2)) + atol
            tol = rtol * esum + atol
            
            if np.any(errs > tol) and verbose:
                # print(errs[errs > tol])
                print(out1[errs > tol])
                print(out2[errs > tol])
                print(tol[errs > tol])
            
            if assertion:
                assert np.all(errs < tol), 'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}'.format(el, errs.max(), errs.mean(), errs.var())
        
            errors.append((el, errs.mean()))
    
        return errors
        
        # init.deltaorthonormal_init(self.weights.data, self.basisexpansion)
        # filter = self.basisexpansion()
        # center = self.s // 2
        # filter = filter[..., center, center]
        # assert torch.allclose(torch.eye(filter.shape[1]), filter.t() @ filter, atol=3e-7)

    def export(self):
        r"""
        Export this module to a normal PyTorch :class:`torch.nn.ConvTranspose2d` module and set to "eval" mode.

        """
    
        # set to eval mode so the filter and the bias are updated with the current
        # values of the weights
        self.eval()
        _filter = self.filter
        _bias = self.expanded_bias
    
        # build the PyTorch Conv2d module
        has_bias = self.bias is not None
        conv = torch.nn.ConvTranspose2d(self.in_type.size,
                                       self.out_type.size,
                                       self.kernel_size,
                                       padding=self.padding,
                                       stride=self.stride,
                                       dilation=self.dilation,
                                       groups=self.groups,
                                       bias=has_bias)
        
        # set the filter and the bias
        conv.weight.data[:] = _filter.data
        if has_bias:
            conv.bias.data[:] = _bias.data
    
        return conv





from escnn.nn import GeometricTensor
from escnn.nn import FieldType
from .equivariant_module import EquivariantModule

from typing import Tuple

import torch

import math

__all__ = ["MaskModule"]


def build_mask(s, margin=2, dtype=torch.float32):
    mask = torch.zeros(1, 1, s, s, dtype=dtype)
    c = (s-1) / 2
    t = (c - margin/100.*c)**2
    sig = 2.
    for x in range(s):
        for y in range(s):
            r = (x - c) ** 2 + (y - c) ** 2
            if r > t:
                mask[..., x, y] = math.exp((t - r)/sig**2)
            else:
                mask[..., x, y] = 1.
    return mask


class MaskModule(EquivariantModule):

    def __init__(self, in_type: FieldType, S: int, margin: float = 0.):
        r"""
        
        Performs an element-wise multiplication of the input with a *mask* of shape ``S x S``.
        
        The mask has value :math:`1` in all pixels with distance smaller than ``(S-1)/2 * (1 - margin)/100`` from the
        center of the mask and :math:`0` elsewhere. Values change smoothly between the two regions.
        
        This operation is useful to remove from an input image or feature map all the part of the signal defined on the
        pixels which lay outside the circle inscribed in the grid.
        Because a rotation would move these pixels outside the grid, this information would anyways be
        discarded when rotating an image. However, allowing a model to use this information might break the guaranteed
        equivariance as rotated and non-rotated inputs have different information content.
        
        
        .. note::
        
            In order to perform the masking, the module expects an input with the same spatial dimensions as the mask.
            Then, input tensors must have shape ``B x C x S x S``.
        
        
        Args:
            in_type (FieldType): input field type
            S (int): the shape of the mask and the expected inputs
            margin (float, optional): margin around the mask in percentage with respect to the radius of the mask
                                      (default ``0.``)
        
        """
        super(MaskModule, self).__init__()

        self.margin = margin
        self.mask = torch.nn.Parameter(build_mask(S, margin=margin), requires_grad=False)

        self.in_type = self.out_type = in_type

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        assert input.type == self.in_type

        assert input.tensor.shape[2:] == self.mask.shape[2:]

        out = input.tensor * self.mask
        return GeometricTensor(out, self.out_type, input.coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        return input_shape






import numpy as np
import torch

from .steerable_filters_basis import SteerableFiltersBasis


from escnn.group import *

from typing import Type, Union, Tuple, Dict, List, Iterable, Callable, Set


class SparseOrbitBasis(SteerableFiltersBasis):
    
    def __init__(self,
                 X: HomSpace,
                 action: Representation,
                 root: np.ndarray,
                 sigma: float,
                 harmonics: List = None,
                 change_of_basis: np.ndarray = None,
                 attributes: Dict = None):
        r"""

        Class which implements a steerable basis for scalar functions on an Euclidean space :math:`\R^n` by considering
        a single *orbit* of the symmetry group :math:`G` in :math:`\R^n`.

        Indeed, an orbit :math:`\{g.x | g \in G\}`, with :math:`x \in \R^n`, is isomorphic to
        an homogeneous space ```X``` of :math:`G`.
        A steerable basis for scalar functions over an homogeneous space is given by the harmonic basis in
        :class:`~escnn.group.HomSpace`.

        In this class, a steerable basis over :math:`\R^n` is defined by embedding the finite number of points of
        a *finite* homogeneous space ```X``` into :math:`\R^n` and then "diffusing" the harmonic basis defined over
        these points in the ambient space using a Gaussian kernel.

        This class only supports finite homogeneous spaces generated by a finite symmetry group, i.e. not only
        :math:`X=G/H` must have a finite number of elements but also :math:`G` itself should.

        The embedding of ```X``` into :math:`\R^n` is defined via the point ``root``, which defines the embedding of the
        coset :math:`eH` associated with the identity, and the ``action`` of the group :math:`G` on :math:`\R^n`.
        This implies that the action of any element of :math:`H` should keep ``root`` fixed.

        Args:
            X (HomSpace): the homogeneous space isomorphic to the orbit in the Euclidean space
            action (Representation): the action of the group ``G`` on the Euclidean space
            root (~np.ndarray): the embedding of the identity coset
            sigma (float): the standard deviation of the Gaussian kernel used to diffuse the harmonic basis in the
                           ambient space.
            harmonics (list, optional): optionally, select only a subset of the harmonic functions. The list should
                                        contain the ids of group's irreps which should be used to construct the
                                        harmonic basis. See also :meth:`~escnn.group.HomSpace.scalar_basis`.
            change_of_basis (~np.ndarray, optional): optionally, apply a transformation on the embedded points to
                                                     construct a different basis. Equivalently, one can apply this
                                                     change of basis to both ``root`` and ``action``.
            attributes (dict, optional): additional attributes to describe the generated basis.

        Attributes:
            ~.X (HomSpace): the homogeneous space

        """

        self.X: HomSpace = X

        # check that the homogeneous space has a finite number of elements
        # unfortunately, we can not directly check the size of the homogeneous space G/H so we check that G itself has
        # only a finite number of elements
        assert self.X.G.order() > 0

        if harmonics is None:
            harmonics = [psi.id for psi in self.X.G.irreps()]

        js = [
            (j, self.X.dimension_basis(j, self.X.H.trivial_representation.id)[1])
            for j in harmonics
            if self.X.dimension_basis(j, self.X.H.trivial_representation.id)[1] > 0
        ]

        super(SparseOrbitBasis, self).__init__(self.X.G, action, js)

        assert root.shape == (self.dimensionality,), (root.shape, self.dimensionality)

        self.root = root
        self.action = action

        root = root.reshape(-1, 1)

        for h in self.X.H.elements:
            assert np.allclose(root, action(self.X._inclusion(h)) @ root, atol=1e-5, rtol=1e-3)

        points = np.concatenate([
            action(g) @ root for g in self.X.G.elements
        ], axis=1)
        _, idx = np.unique(
            np.round(points, decimals=4),
            axis=1, return_index=True
        )
        points = points[:, idx].T

        assert points.shape == (self.X.G.order() / self.X.H.order(), self.dimensionality)

        assert sigma > 0., sigma

        for j, m in self.js:

            _harmonics_j = torch.tensor(np.stack([
                self.X.scalar_basis(self.X.G.elements[i], j)[..., 0].T
                for i in idx
            ], axis=0), dtype=torch.float32)
            assert _harmonics_j.shape == (idx.shape[0], m, self.X.G.irrep(*j).size), (_harmonics_j.shape, m, j)

            self.register_buffer(f'harmonics_{j}', _harmonics_j)

        self.sigma = sigma

        points = torch.tensor(points, dtype=torch.float32)
        if change_of_basis is not None:
            assert change_of_basis.shape == (self.dimensionality, self.dimensionality)
            # check that the matrix is invertible
            assert np.isfinite(np.linalg.cond(change_of_basis))

            change_of_basis = torch.tensor(change_of_basis, dtype=points.dtype, device=points.device)

            points = points @ change_of_basis.T

        self.register_buffer('points', points)

        self.change_of_basis = change_of_basis

        self._attributes = attributes if attributes is not None else dict()

    def _get_harmonics(self, j):
        return getattr(self, f'harmonics_{j}')

    def sample(self, points: torch.Tensor, out: torch.Tensor = None) -> torch.Tensor:
        r"""

        Sample the continuous basis elements on the discrete set of points in ``points``.
        Optionally, store the resulting multidimentional array in ``out``.

        ``points`` must be an array of shape `(N, d)`, where `N` is the number of points and
        `d` is the :meth:`~escnn.kernels.SteerableFilterBasis.dimensionality` of the Euclidean space.

        Args:
            points (~torch.Tensor): points where to evaluate the basis elements
            out (~torch.Tensor, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """
    
        assert len(points.shape) == 2
        S = points.shape[0]
        assert points.shape[1] == self.dimensionality
    
        if out is None:
            out = torch.empty((S, self.dim, 1, 1), dtype=self.points.dtype, device=self.points.device)
    
        assert out.shape == (S, self.dim, 1, 1)

        assert self.points.device == points.device, (self.points.device, points.device)
        assert out.device == points.device, (out.device, points.device)

        weights = self.points.unsqueeze(1) - points.unsqueeze(0)
        assert weights.shape == (self.points.shape[0], S, self.dimensionality)

        weights = (weights**2).sum(axis=2) / self.sigma**2
        weights = torch.exp(- 0.5 * weights)

        B = 0
        for j, m in self.js:
            out[:, B:B + self.dim_harmonic(j), ...].view(
                S, m, -1
            )[:] = torch.einsum('irm,io->orm', self._get_harmonics(j), weights)
            
            B += self.dim_harmonic(j)

        return out

    def __iter__(self) -> Iterable:

        idx = 0
        for j, M in self.js:
            psi = self.group.irrep(*j)

            attr = {
                'irrep:' + k: v
                for k, v in psi.attributes.items()
            }
            attr.update(**self._attributes)
            attr["j"] = j
            attr["shape"] = (1, 1)

            for i in range(M):
                for m in range(psi.size):
                    _attr = {"idx" : idx, "i" : i, "m" : m}
                    _attr.update(**attr)
                    yield _attr
                    idx += 1

    def __getitem__(self, idx) -> Dict:
        assert idx < self.dim, (idx, self.dim)

        j, j_idx = self._get_j_from_idx(idx)

        attr = {
            'irrep:' + k: v
            for k, v in self.group.irrep(*j).attributes.items()
        }
        attr.update(**self._attributes)

        attr["idx"] = idx
        attr["j"] = j
        attr["i"] = j_idx // self.group.irrep(*j).size
        attr["m"] = j_idx % self.group.irrep(*j).size
        attr["shape"] = (1, 1)

        return attr

    def steerable_attrs_j_iter(self, j: Tuple) -> Iterable:
        attr = {
            'irrep:' + k: v
            for k, v in self.X.G.irrep(*j).attributes.items()
        }
        attr['sigma'] = self.sigma
        attr.update(**self._attributes)

        for idx in range(self.dim_harmonic(j)):
            _attr = {'i': idx}
            _attr.update(**attr)
            yield _attr

    def steerable_attrs_j(self, j: Tuple, idx) -> Dict:

        assert 0 <= idx < self.dim_harmonic(j), idx

        attr = {
            'irrep:' + k: v
            for k, v in self.X.G.irrep(*j).attributes.items()
        }

        attr['i'] = idx
        attr['sigma'] = self.sigma
        attr.update(**self._attributes)
        return attr

    def __eq__(self, other):
        if not isinstance(other, SparseOrbitBasis):
            return False
        elif self.X != other.X:
            return False
        elif self.action != other.action:
            return False
        elif not np.allclose(self.root, other.root, atol=1e-5):
            return False
        elif self.js != other.js:
            return False
        elif not np.isclose(self.sigma, other.sigma):
            return False
        elif (self.change_of_basis is None) != (other.change_of_basis is None):
            return False
        elif self.change_of_basis is not None:
            return torch.allclose(self.change_of_basis, other.change_of_basis)
        else:
            return True

    def __hash__(self):
        return 100 * hash(self.X) + 10 * hash(self.root.tobytes()) + 10 * hash(self.sigma) + hash(str(self.change_of_basis)) + sum(hash(x) for x in self.js) + hash(self.action)


class SparseOrbitBasisWithIcosahedralSymmetry(SparseOrbitBasis):

    def __init__(self,
                 X: HomSpace,
                 sigma: float,
                 harmonics: List = None,
                 change_of_basis: np.ndarray = None,
                 attributes: Dict = None):
        r"""

        Given an homogeneous space ``X`` of the Icosahedral group, builds the :class:`~escnn.kernels.SparseOrbitBasis`
        with a ``root`` which generates an orbit isomorphic to ``X``.

        """

        G = ico_group()

        assert X.G == G, X.G

        assert X.sgid[:2] in [(False, 2), (False, 3), (False, 5)], X.sgid
        if len(X.sgid) > 2:
            assert len(X.sgid) == 3
            assert np.allclose(np.fabs(X.sgid[2].value), np.array([0., 0., 0., 1.]))

        # retrieve the rotation axis of one of the elements of the stabilizer group to find the point in R^3 which
        # represents the coset of the identity element of G
        axis = X._inclusion(X.H.elements[1]).to('Q')[:3]
        root = axis.reshape(-1) / np.linalg.norm(axis)

        super(SparseOrbitBasisWithIcosahedralSymmetry, self).__init__(
            X,
            G.standard_representation,
            root,
            sigma, harmonics, change_of_basis, attributes
        )


from __future__ import annotations

import escnn.group
from escnn.group import Group, GroupElement

from ._numerical import decompose_representation_finitegroup

from collections import defaultdict

from typing import Callable, Any, List, Tuple, Dict, Union, Set

import math
import numpy as np
import scipy as sp
from scipy import linalg, sparse
from scipy.sparse.csgraph import connected_components


__all__ = [
    "Representation",
    "build_from_discrete_group_representation",
    "directsum",
    "disentangle",
    "change_basis",
    "build_regular_representation",
    "build_quotient_representation",
    "build_induced_representation",
    "homomorphism_space"
]


class Representation:
    
    def __init__(self,
                 group: Group,
                 name: str,
                 irreps: List[Tuple],
                 change_of_basis: np.ndarray,
                 supported_nonlinearities: Union[List[str], Set[str]] = [],
                 representation: Union[Dict[GroupElement, np.ndarray], Callable[[GroupElement], np.ndarray]] = None,
                 character: Union[Dict[GroupElement, float], Callable[[Any], float]] = None,
                 change_of_basis_inv: np.ndarray = None,
                 **kwargs):
        r"""
        Class used to describe a group representation.
        
        A (real) representation :math:`\rho` of a group :math:`G` on a vector space :math:`V=\mathbb{R}^n` is a map
        (a *homomorphism*) from the group elements to invertible matrices of shape :math:`n \times n`, i.e.:
        
        .. math::
            \rho : G \to \GL{V}
            
        such that the group composition is modeled by a matrix multiplication:
        
        .. math::
            \rho(g_1 g_2) = \rho(g_1) \rho(g_2) \qquad  \forall \ g_1, g_2 \in G \ .
        
        Any representation (of a compact group) can be decomposed into the *direct sum* of smaller, irreducible
        representations (*irreps*) of the group up to a change of basis:
        
        .. math::
            \forall \ g \in G, \ \rho(g) = Q \left( \bigoplus\nolimits_{i \in I} \psi_i(g) \right) Q^{-1} \ .
        
        Here :math:`I` is an index set over the irreps of the group :math:`G` which are contained in the
        representation :math:`\rho`.
        
        This property enables one to study a representation by its irreps and it is used here to work with arbitrary
        representations.
        
        :attr:`escnn.group.Representation.change_of_basis` contains the change of basis matrix :math:`Q` while
        :attr:`escnn.group.Representation.irreps` is an ordered list containing the names of the irreps :math:`\psi_i`
        indexed by the index set :math:`I`.
        
        A ``Representation`` instance can be used to describe a feature field in a feature map.
        It is the building block to build the representation of a feature map, by "stacking" multiple representations
        (taking their *direct sum*).
        
        .. note ::
            In most of the cases, it should not be necessary to manually instantiate this class.
            Indeed, the user can build the most common representations or some custom representations via the following
            methods and functions:
            
            - :meth:`escnn.group.Group.irrep`,
            
            - :meth:`escnn.group.Group.regular_representation`,
            
            - :meth:`escnn.group.Group.quotient_representation`,
            
            - :meth:`escnn.group.Group.induced_representation`,
            
            - :meth:`escnn.group.Group.restrict_representation`,
            
            - :meth:`escnn.group.Representation.tensor`,
            
            - :func:`escnn.group.directsum`,
            
            - :func:`escnn.group.change_basis`
            
            Additionally, the direct sum of two representations can be quickly generated using the binary operator `+`,
            see :meth:`~escnn.group.Representation.__add__`.
            
            
        
        If ``representation`` is ``None`` (default), it is automatically inferred by evaluating each irrep, stacking
        their results (through direct sum) and then applying the changes of basis. **Warning**: the representation of an
        element is built at run-time every time this object is called (through ``__call__``), so this approach might
        become computationally expensive with large representations.
        
        Analogously, if the ``character`` of the representation is ``None`` (default), it is automatically inferred
        evaluating ``representation`` and computing its trace.
        
        .. note ::
            It is assumed that both ``representation`` and ``character`` expect input group elements in the default
            parametrization of ``group``, i.e. :attr:`escnn.group.Group.PARAM`.
        
        .. todo::
            improve the interface for "supported non-linearities" and write somewhere the available options
        
        Args:
            group (Group): the group to be represented.
            name (str): an identification name for this representation.
            irreps (list): a list of irreps' ids. Each id is a tuple representing one of the *irreps* of the
                    group (see :attr:`escnn.group.Group.irreps` and :attr:`escnn.group.IrreducibleRepresentation.id`).
            change_of_basis (~numpy.ndarray, optional): the matrix which transforms the direct sum of the irreps
                    in this representation. By default (`None`), the identity is assumed.
            supported_nonlinearities (list or set, optional): a list or set of nonlinearity types supported by this
                    representation.
            representation (dict or callable, optional): a callable implementing this representation or a dict mapping
                    each group element to its representation.
            character (callable or dict, optional): a callable returning the character of this representation for an
                    input element or a dict mapping each group element to its character.
            change_of_basis_inv (~numpy.ndarray, optional): the inverse of the ``change_of_basis`` matrix; if not
                    provided (``None``), it is computed from ``change_of_basis``.
            **kwargs: custom attributes the user can set and, then, access from the dictionary in
                    :attr:`escnn.group.Representation.attributes`
            
        Attributes:
            ~.group (Group): The group which is being represented.
            ~.name (str): A string identifying this representation.
            ~.size (int): Dimensionality of the vector space of this representation. In practice, this is the size of the
                matrices this representation maps the group elements to.
            ~.change_of_basis (~numpy.ndarray): Change of basis matrix for the irreps decomposition.
            ~.change_of_basis_inv (~numpy.ndarray): Inverse of the change of basis matrix for the irreps decomposition.
            ~.representation (callable): Method implementing the map from group elements to their representation matrix.
            ~.supported_nonlinearities (set): A set of strings identifying the non linearities types supported by this representation.
            ~.irreps (list): List of irreps into which this representation decomposes.
            ~.attributes (dict): Custom attributes set when creating the instance of this class.
            ~.irreducible (bool): Whether this is an irreducible representation or not (i.e. if it can't be decomposed into further invariant subspaces).

        
        """
        
        # can't have the name of an already existing representation
        assert name not in group.representations, f"A representation for {group.name} with name {name} already exists!"

        assert len(change_of_basis.shape) == 2 and change_of_basis.shape[0] == change_of_basis.shape[1], change_of_basis.shape
        assert change_of_basis.shape[0] > 0, change_of_basis.shape

        if change_of_basis_inv is None:
            change_of_basis_inv = sp.linalg.inv(change_of_basis)

        assert len(change_of_basis_inv.shape) == 2
        assert change_of_basis_inv.shape[0] == change_of_basis.shape[0]
        assert change_of_basis_inv.shape[1] == change_of_basis.shape[1]
        assert np.allclose(change_of_basis @ change_of_basis_inv, np.eye(change_of_basis.shape[0]))
        assert np.allclose(change_of_basis_inv @ change_of_basis, np.eye(change_of_basis.shape[0]))
        
        # Group: A string identifying this representation.
        self.group = group
        
        # str: The group this is a representation of.
        self.name = name
        
        # int: Dimensionality of the vector space of this representation.
        # In practice, this is the size of the matrices this representation maps the group elements to.
        self.size = change_of_basis.shape[0]

        if len(irreps) > 1:
            irreps_size = sum(self.group.irrep(*irr).size for irr in irreps)
            assert irreps_size == self.size, \
                f"Error! The size of the change of basis ({self.size}) does not match the sum of the sizes of the irreps ({irreps_size})."
        
        # np.ndarray: Change of basis matrix for the irreps decomposition.
        self.change_of_basis = change_of_basis

        # np.ndarray: Inverse of the change of basis matrix for the irreps decomposition.
        self.change_of_basis_inv = change_of_basis_inv
        
        # list(tuple): List of irreps this representation decomposes into
        self.irreps = irreps

        if representation is None:
            irreps_instances = [group.irrep(*n) for n in irreps]
            if np.allclose(change_of_basis, np.eye(self.size)):
                representation = direct_sum_factory(irreps_instances, None, None)
            else:
                representation = direct_sum_factory(irreps_instances, change_of_basis, change_of_basis_inv)
        elif isinstance(representation, dict):
            assert set(representation.keys()) == set(self.group.elements), "Error! Keys don't match group's elements"
            
            self._stored_representations = representation
            representation = _build_representation_callable_from_dict(self._stored_representations)
            
        elif not callable(representation):
            raise ValueError('Error! "representation" is neither a dictionary nor callable')
        
        # Callable: Method implementing the map from group elements to matrix representations.
        self.representation = representation

        if isinstance(character, dict):
            
            assert set(character.keys()) == set(self.group.elements), "Error! Keys don't match group's elements"
            
            self._characters = character

        elif callable(character):
            self._characters = character
        elif character is None:
            # if the character is not given as input, it is automatically inferred from the given representation
            # taking its trace
            self._characters = None
        else:
            raise ValueError('Error! "character" must be a dictionary, a callable or "None"')

        # TODO - assert size matches size of the matrix returned by the callable
        
        self.supported_nonlinearities = set(supported_nonlinearities)
        
        # dict: Custom attributes set when creating the instance of this class
        self.attributes = kwargs

        # TODO : remove the condition of an identity change of basis?
        # bool: Whether this is an irreducible representation or not (i.e.: if it can't be decomposed further)
        self.irreducible = len(self.irreps) == 1 and np.allclose(self.change_of_basis, np.eye(self.change_of_basis.shape[0]))

        self._cached_restricted_representations = dict()
        
        self._irreps_multiplicities = defaultdict(int)
        
        for irrep in self.irreps:
            self._irreps_multiplicities[irrep] += 1
    
    def character(self, e: escnn.group.GroupElement) -> float:
        r"""

        The *character* of a finite-dimensional real representation is a function mapping a group element
        to the trace of its representation:

        .. math::

            \chi_\rho: G \to \mathbb{C}, \ \ g \mapsto \chi_\rho(g) := \operatorname{tr}(\rho(g))

        It is useful to perform the irreps decomposition of a representation using *Character Theory*.
        
        Args:
            e (GroupElement): an element of the group of this representation
        
        Returns:
            the character of the element
        
        """

        assert e.group == self.group, f"Error: the element {e} belongs to the group {e.group.name} and not {self.group.name}"
        
        if self._characters is None:
            # if the character is not given as input, it is automatically inferred from the given representation
            # taking its trace
            repr = self(e)
            return np.trace(repr)
        elif isinstance(self._characters, dict):
            return self._characters[e]
        elif callable(self._characters):
            return self._characters(e)
        else:
            raise RuntimeError('Error! Character not recognized!')

    def is_trivial(self) -> bool:
        r"""
        
        Whether this representation is trivial or not.
        
        Returns:
            if the representation is trivial

        """
        return self.irreducible and self.group.trivial_representation.id == self.irreps[0]
    
    def contains_trivial(self) -> bool:
        r"""

        Whether this representation contains the trivial representation among its irreps.
        This is an alias for::
            
            any(self.group.irreps[irr].is_trivial() for irr in self.irreps)

        Returns:
           if it contains the trivial representation

        """
        for irrep in self.irreps:
            if self.group.irrep(*irrep).is_trivial():
                return True
        return False

    def restrict(self, id) -> escnn.group.Representation:
        r"""
        
        Restrict the current representation to the subgroup identified by ``id``.
        Check the documentation of the :meth:`~escnn.group.Group.subgroup` method in the underlying group to see the
        available subgroups and accepted ids.

        .. note ::
            This operation is cached.
            Multiple calls using the same subgroup ``id`` will return the same instance instead of computing a new
            restriction.

        Args:
            id: identifier of the subgroup

        Returns:
            the restricted representation
        """

        if id not in self._cached_restricted_representations:
            self._cached_restricted_representations[id] = self.group.restrict_representation(id, self)

        return self._cached_restricted_representations[id]
    
    def __call__(self, element: GroupElement) -> np.ndarray:
        """
        An instance of this class can be called and it implements the mapping from an element of a group to its
        representation.
        
        This is equivalent to calling :meth:`escnn.group.Representation.representation`,
        though ``__call__`` first checks ``element`` is a valid input (i.e. an element of the group).
        It is recommended to use this call.

        Args:
            element (GroupElement): an element of the group

        Returns:
            A matrix representing the input element

        """
        
        assert element.group == self.group, f"Error: the element {element} belongs to the group {element.group.name} and not {self.group.name}"
        
        return self.representation(element)

    def __add__(self, other: escnn.group.Representation) -> escnn.group.Representation:
        r"""

        Compute the *direct sum* of two representations of a group.

        The two representations need to belong to the same group.

        Args:
            other (Representation): another representation

        Returns:
            the direct sum

        """
        
        return directsum([self, other])
    
    def __eq__(self, other: escnn.group.Representation) -> bool:
        if not isinstance(other, Representation):
            return False
        
        return (self.name == other.name
                and self.group == other.group
                and np.allclose(self.change_of_basis, other.change_of_basis)
                and self.irreps == other.irreps
                and self.supported_nonlinearities == other.supported_nonlinearities)
    
    def __repr__(self) -> str:
        return f"{self.group.name}|[{self.name}]:{self.size}"
    
    def __hash__(self):
        return hash(repr(self))
    
    def tensor(self, other: Representation) -> Representation:
        r"""
            Compute the tensor product representation of the current representation and the input representaton.
        """
        return self.group._tensor_product(self, other)
    
    def endomorphism_basis(self) -> np.ndarray:
        r"""
            Compute a basis for the space of endomorphism of this representation from the endomorphism spaces of the
            irreps contained in this representation.
            
            .. warning::
                This basis might become quite big for large representations and, therefore, this method might get
                computationally expensive.
                In practice, we recommend using this method only to generate the basis for small representations (e.g.
                for irreps).
                This basis is generated and stored more efficiently in the :doc:`escnn.nn` package, when parameterizing
                neural network modules; e.g. see :class:`~escnn.nn.BlocksBasisExpansion`.
            
        """

        dim = 0
        for in_irrep in self.irreps:
            in_irrep = self.group.irrep(*in_irrep)
            for out_irrep in self.irreps:
                out_irrep = self.group.irrep(*out_irrep)
                if in_irrep == out_irrep:
                    dim += in_irrep.sum_of_squares_constituents

        basis = np.zeros((dim, self.size, self.size))
    
        p = 0
        i = 0
        for in_irrep in self.irreps:
            in_irrep = self.group.irrep(*in_irrep)
            
            j = 0
            for out_irrep in self.irreps:
                out_irrep = self.group.irrep(*out_irrep)
                
                if out_irrep == in_irrep:
    
                    basis[p:p+in_irrep.sum_of_squares_constituents, j:j+in_irrep.size, i:i+in_irrep.size] = in_irrep.endomorphism_basis()
                    
                    p += in_irrep.sum_of_squares_constituents
                    
                j += out_irrep.size
            i += in_irrep.size

        basis = np.einsum('Mm,kmn,nN->kMN', self.change_of_basis, basis, self.change_of_basis_inv)
        return basis
    
    def multiplicity(self, irrep: tuple) -> int:
        r"""
            Returns the multiplicity of the ``irrep`` in the current representation.
        """
        return self._irreps_multiplicities[irrep]


def _build_representation_callable_from_dict(repr_dict: Dict[GroupElement, np.ndarray]):
    
    def representation(e: GroupElement, repr_dict: Dict = repr_dict) -> np.ndarray:
        return repr_dict[e]
    
    return representation


def _build_character_from_irreps_multiplicities(irreps: List[Tuple['IrreducibleRepresentation', int]]):
    
    def character(element: GroupElement, irreps=irreps) -> float:
        return sum([m * irrep.character(element) for (irrep, m) in irreps])

    return character


def build_from_discrete_group_representation(representation: Dict[GroupElement, np.array],
                                             name: str,
                                             group: escnn.group.Group,
                                             supported_nonlinearities: List[str]
                                             ) -> escnn.group.Representation:
    r"""
    Given a representation of a *finite* group as a dictionary of matrices, the method decomposes it as a direct sum
    of the irreps of the group and computes the change-of-basis matrix. Then, a new instance of
    :class:`~escnn.group.Representation` is built using the direct sum of irreps and the change-of-basis matrix as
    representation taking as input elements from the continuous parent group.
    
    For instance, given a regular representation of a cyclic group of order :math:`n` implemented as a
    list of permutations matrices, the method builds a representation of SO(2) whose values are these permutation
    matrices when evaluated to the angles corresponding to the elements of the cyclic group (i.e. any angle in the
    form :math:`k 2 \pi / n` with :math:`k` in :math:`[0, \dots, n-1]`)
    
    Args:
        representation (dict): a dictionary mapping an element of ``group`` to a numpy array (must be a squared matrix)
        name (str): an identification name of the representation
        group (Group): the group whose representation has to be built
        supported_nonlinearities (list): list of non linearities types supported by this representation.
    
    Returns:
        a new representation
        
    """

    assert group.order() > 0

    assert set(representation.keys()) == set(group.elements), "Error! Keys don't match group's elements"

    # decompose the representation
    cob, multiplicities = decompose_representation_finitegroup(representation, group)

    # build a list of representation instances with their multiplicities
    irreps_with_multiplicities = [(group.irrep(*id), m) for (id, m) in multiplicities]

    # build the character of this representation
    new_character = _build_character_from_irreps_multiplicities(irreps_with_multiplicities)

    irreps = []
    for irr, m in multiplicities:
       irreps += [irr] * m

    # build the representation object
    return Representation(group,
                          name,
                          irreps,
                          cob,
                          supported_nonlinearities,
                          representation=representation,
                          character=new_character)


# TODO when built from "directsum" we can "optimize" the representation by sorting the internal irreps
#      and permuting the change of basis matrix's columns accordingly. Could be useful when one uses GNORM batchnorm

def directsum(reprs: List[escnn.group.Representation],
              change_of_basis: np.ndarray = None,
              name: str = None
              ) -> escnn.group.Representation:
    r"""

    Compute the *direct sum* of a list of representations of a group.
    
    The direct sum of two representations is defined as follow:
    
    .. math::
        \rho_1(g) \oplus \rho_2(g) = \begin{bmatrix} \rho_1(g) & 0 \\ 0 & \rho_2(g) \end{bmatrix}
    
    This can be generalized to multiple representations as:
    
    .. math::
        \bigoplus_{i=1}^I \rho_i(g) = (\rho_1(g) \oplus (\rho_2(g) \oplus (\rho_3(g) \oplus \dots = \begin{bmatrix}
            \rho_1(g) &         0 &  \dots &      0 \\
                    0 & \rho_2(g) &  \dots & \vdots \\
               \vdots &    \vdots & \ddots &      0 \\
                    0 &     \dots &      0 & \rho_I(g) \\
        \end{bmatrix}
    

    .. note::
        All the input representations need to belong to the same group.

    Args:
        reprs (list): the list of representations to sum.
        change_of_basis (~numpy.ndarray, optional): an invertible square matrix to use as change of basis after computing the direct sum.
                By default (``None``), an identity matrix is used, such that only the direct sum is evaluated.
        name (str, optional): a name for the new representation.

    Returns:
        the direct sum

    """
    
    group = reprs[0].group
    for r in reprs:
        assert group == r.group
    
    if name is None:
        name = "_".join([f"[{r.name}]" for r in reprs])
    
    irreps = []
    for r in reprs:
        irreps += r.irreps
    
    size = sum([r.size for r in reprs])
    
    cob = np.zeros((size, size))
    cob_inv = np.zeros((size, size))
    p = 0
    for r in reprs:
        cob[p:p + r.size, p:p + r.size] = r.change_of_basis
        cob_inv[p:p + r.size, p:p + r.size] = r.change_of_basis_inv
        p += r.size

    if change_of_basis is not None:
        change_of_basis = change_of_basis @ cob
        change_of_basis_inv = sp.linalg.inv(change_of_basis)
    else:
        change_of_basis = cob
        change_of_basis_inv = cob_inv

    supported_nonlinearities = set.intersection(*[r.supported_nonlinearities for r in reprs])
    
    return Representation(group, name, irreps, change_of_basis, supported_nonlinearities, change_of_basis_inv=change_of_basis_inv)


def disentangle(repr: Representation) -> Tuple[np.ndarray, List[Representation]]:
    r"""
    
    If possible, disentangle the input representation by decomposing it into the direct sum of smaller representations
    and a change of basis acting as a permutation matrix.
    
    This method is useful to decompose a feature vector transforming with a complex representation into multiple feature
    vectors which transform independently with simpler representations.
    
    Note that this method only decomposes a representation by applying a permutation of axes.
    A more general decomposition using any invertible matrix is possible but is just a decomposition into
    irreducible representations (see :class:`~escnn.group.Representation`).
    However, since the choice of change of basis is relevant for the kind of operations which can be performed
    (e.g. non-linearities), it is often not desirable to discard any change of basis and completely disentangle a
    representation.
    
    Considering only change of basis matrices which are permutation matrices is sometimes more useful.
    For instance, the restriction of the regular representation of a group to a subgroup results in a representation containing
    multiple regular representations of the subgroup (one for each `coset`).
    However, depending on how the original representation is built, the restricted representation might not be
    block-diagonal and, so, the subgroup's regular representations might not be clearly separated.
    
    For example, this happens when restricting the regular representation of :math:`\D3`
    
    +-----------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
    |    :math:`g`                      |          :math:`e`                                                                                                                                                                       |          :math:`r`                                                                                                                                                                       |        :math:`r^2`                                                                                                                                                                       |          :math:`f`                                                                                                                                                                       |         :math:`rf`                                                                                                                                                                       |       :math:`r^2f`                                                                                                                                                                       |
    +===================================+==========================================================================================================================================================================================+==========================================================================================================================================================================================+==========================================================================================================================================================================================+==========================================================================================================================================================================================+==========================================================================================================================================================================================+==========================================================================================================================================================================================+
    |  :math:`\rho_\text{reg}^{\D3}(g)` | :math:`\begin{bmatrix} 1 & 0 & 0 & 0 & 0 & 0 \\ 0 & 1 & 0 & 0 & 0 & 0 \\ 0 & 0 & 1 & 0 & 0 & 0 \\ 0 & 0 & 0 & 1 & 0 & 0 \\ 0 & 0 & 0 & 0 & 1 & 0 \\ 0 & 0 & 0 & 0 & 0 & 1 \end{bmatrix}` | :math:`\begin{bmatrix} 0 & 0 & 1 & 0 & 0 & 0 \\ 1 & 0 & 0 & 0 & 0 & 0 \\ 0 & 1 & 0 & 0 & 0 & 0 \\ 0 & 0 & 0 & 0 & 0 & 1 \\ 0 & 0 & 0 & 1 & 0 & 0 \\ 0 & 0 & 0 & 0 & 1 & 0 \end{bmatrix}` | :math:`\begin{bmatrix} 0 & 1 & 0 & 0 & 0 & 0 \\ 0 & 0 & 1 & 0 & 0 & 0 \\ 1 & 0 & 0 & 0 & 0 & 0 \\ 0 & 0 & 0 & 0 & 1 & 0 \\ 0 & 0 & 0 & 0 & 0 & 1 \\ 0 & 0 & 0 & 1 & 0 & 0 \end{bmatrix}` | :math:`\begin{bmatrix} 0 & 0 & 0 & 1 & 0 & 0 \\ 0 & 0 & 0 & 0 & 0 & 1 \\ 0 & 0 & 0 & 0 & 1 & 0 \\ 1 & 0 & 0 & 0 & 0 & 0 \\ 0 & 0 & 1 & 0 & 0 & 0 \\ 0 & 1 & 0 & 0 & 0 & 0 \end{bmatrix}` | :math:`\begin{bmatrix} 0 & 0 & 0 & 0 & 1 & 0 \\ 0 & 0 & 0 & 1 & 0 & 0 \\ 0 & 0 & 0 & 0 & 0 & 1 \\ 0 & 1 & 0 & 0 & 0 & 0 \\ 1 & 0 & 0 & 0 & 0 & 0 \\ 0 & 0 & 1 & 0 & 0 & 0 \end{bmatrix}` | :math:`\begin{bmatrix} 0 & 0 & 0 & 0 & 0 & 1 \\ 0 & 0 & 0 & 0 & 1 & 0 \\ 0 & 0 & 0 & 1 & 0 & 0 \\ 0 & 0 & 1 & 0 & 0 & 0 \\ 0 & 1 & 0 & 0 & 0 & 0 \\ 1 & 0 & 0 & 0 & 0 & 0 \end{bmatrix}` |
    +-----------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
    
    to the reflection group :math:`\C2`
    
    +--------------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
    |    :math:`g`                                     |          :math:`e`                                                                                                                                                                       |          :math:`f`                                                                                                                                                                       |
    +==================================================+==========================================================================================================================================================================================+==========================================================================================================================================================================================+
    |  :math:`\Res{\C2}{\D3} \rho_\text{reg}^{\D3}(g)` | :math:`\begin{bmatrix} 1 & 0 & 0 & 0 & 0 & 0 \\ 0 & 1 & 0 & 0 & 0 & 0 \\ 0 & 0 & 1 & 0 & 0 & 0 \\ 0 & 0 & 0 & 1 & 0 & 0 \\ 0 & 0 & 0 & 0 & 1 & 0 \\ 0 & 0 & 0 & 0 & 0 & 1 \end{bmatrix}` | :math:`\begin{bmatrix} 0 & 0 & 0 & 1 & 0 & 0 \\ 0 & 0 & 0 & 0 & 0 & 1 \\ 0 & 0 & 0 & 0 & 1 & 0 \\ 1 & 0 & 0 & 0 & 0 & 0 \\ 0 & 0 & 1 & 0 & 0 & 0 \\ 0 & 1 & 0 & 0 & 0 & 0 \end{bmatrix}` |
    +--------------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
    
    Indeed, in :math:`\Res{\C2}{\D3} \rho_\text{reg}^{\D3}(g)` the three pairs of entries (1, 4), (2, 6) and (3, 5)
    never mix with each other but only permute internally.
    Moreover, each pair transform according to the regular representation of :math:`\C2`.
    Through a permutation of the entries, it is possible to make all the entries belonging to the same pair contiguous.
    This this reshuffled representation is then equal to
    :math:`\rho_\text{reg}^{\C2} \oplus \rho_\text{reg}^{\C2} \oplus \rho_\text{reg}^{\C2}`.
    Though theoretically equivalent, an implementation of this representation where the entries are contiguous is
    convenient when computing functions over single fields like batch normalization.
    
    Notice that applying the change of basis returned to the input representation (e.g. through
    :func:`escnn.group.change_basis`) will result in a representation containing the direct sum of the representations
    in the list returned.
    
    .. seealso::
        :func:`~escnn.group.directsum`,
        :func:`~escnn.group.change_basis`
    
    Args:
        repr (Representation): the input representation to disentangle

    Returns:
        a tuple containing
        
            - **change of basis**: a (square) permutation matrix of the size of the input representation
            
            - **representation**: the list of representations the input one is decomposed into
        
    """
    
    rsize = repr.size
    nirreps = len(repr.irreps)
    
    cob_mask = np.isclose(repr.change_of_basis, np.zeros_like(repr.change_of_basis))
    cob_mask = np.invert(cob_mask)
    
    irreps = [repr.group.irrep(*irr) for irr in repr.irreps]
    irreps_pos = np.cumsum([0] + [irr.size for irr in irreps])
    
    masks = []
    i_pos = 0
    for i, irr in enumerate(irreps):
        mask = cob_mask[:, i_pos:i_pos + irr.size].any(axis=1)
        masks.append(mask)
        i_pos += irr.size
    
    cob_mask = np.array(masks, dtype=bool)
    
    graph = np.zeros((nirreps + rsize, nirreps + rsize), dtype=bool)
    graph[:nirreps, nirreps:] = cob_mask
    graph[nirreps:, :nirreps] = cob_mask.T
    
    n_blocks, labels = connected_components(graph, directed=False, return_labels=True)
    
    irreps_labels = labels[:nirreps]
    field_labels = labels[nirreps:]
    
    blocks = [([], []) for _ in range(n_blocks)]
    
    for i in range(nirreps):
        blocks[irreps_labels[i]][0].append(i)
    for i in range(rsize):
        blocks[field_labels[i]][1].append(i)
    
    change_of_basis = np.zeros_like(repr.change_of_basis)
    
    representations = []
    current_position = 0
    for block, (irreps_indices, row_indices) in enumerate(blocks):
        
        irreps_indices = sorted(irreps_indices)
        row_indices = sorted(row_indices)
        
        total_size = len(row_indices)
        
        assert sum([irreps[irr].size for irr in irreps_indices]) == total_size
        
        col_indices = []
        for irr in irreps_indices:
            col_indices += list(range(irreps_pos[irr], irreps_pos[irr]+irreps[irr].size))
        
        assert len(col_indices) == total_size
        
        new_cob = repr.change_of_basis[np.ix_(row_indices, col_indices)]
        
        field_repr = Representation(repr.group,
                                      f"{repr.name}_{block}",
                                      [irreps[id].id for id in irreps_indices],
                                      new_cob,
                                      repr.supported_nonlinearities)
        representations.append(field_repr)

        next_position = current_position + len(row_indices)
        change_of_basis[current_position:next_position, row_indices] = np.eye(len(row_indices))
        
        current_position = next_position
        
    return change_of_basis, representations


def change_basis(repr: Representation,
                 change_of_basis: np.ndarray,
                 name: str,
                 supported_nonlinearities: List[str] = None
                 ) -> Representation:
    r"""
    Build a new representation from an already existing one by applying a change of basis.
    In other words, if :math:`\rho(\cdot)` is the representation and :math:`Q` the change of basis in input, the
    resulting representation will evaluate to :math:`Q \rho(\cdot) Q^{-1}`.
    
    Notice that the change of basis :math:`Q` has to be invertible.
    
    
    Args:
        repr (Representation): the input representation
        change_of_basis (~numpy.ndarray): the change of basis to apply
        name (str, optional): the name to use to identify the new representation
        supported_nonlinearities (list, optional): a list containing the ids of the supported non-linearities
            for the new representation

    Returns:
        the new representation

    """
    assert len(change_of_basis.shape) == 2
    assert change_of_basis.shape[0] == change_of_basis.shape[1]
    assert change_of_basis.shape[0] == repr.size
    
    if supported_nonlinearities is None:
        # by default, no non-linearities are supported
        supported_nonlinearities = []
    
    # compute the new change of basis
    new_cob = change_of_basis @ repr.change_of_basis
    new_cob_inv = repr.change_of_basis_inv @ sp.linalg.inv(change_of_basis)
    
    return Representation(repr.group, name, repr.irreps, new_cob,
                          supported_nonlinearities=supported_nonlinearities,
                          change_of_basis_inv=new_cob_inv)


def build_regular_representation(group: escnn.group.Group) -> Tuple[List[escnn.group.IrreducibleRepresentation], np.ndarray, np.ndarray]:
    r"""
    
    Build the regular representation of the input ``group``.
    As the regular representation has size equal to the number of elements in the group, only
    finite groups are accepted.
    
    Args:
        group (Group): the group whose representations has to be built

    Returns:
        a tuple containing the list of irreps, the change of basis and the inverse change of basis of
        the regular representation

    """
    assert group.order() > 0
    assert group.elements is not None and len(group.elements) > 0
    
    size = group.order()

    index = {e: i for i, e in enumerate(group.elements)}
    
    representation = {}
    character = {}
    
    for e in group.elements:
        # print(index[e], e)
        r = np.zeros((size, size), dtype=float)
        for g in group.elements:
            
            eg = e @ g

            i = index[g]
            j = index[eg]
            
            r[j, i] = 1.0
        
        representation[e] = r
        # the character maps an element to the trace of its representation
        character[e] = np.trace(r)

    # compute the multiplicities of the irreps from the dot product between
    # their characters and the character of the representation
    irreps = []
    multiplicities = []
    for irrep in group.irreps():
        # for each irrep
        multiplicity = 0.0
    
        # compute the inner product with the representation's character
        for element, char in character.items():
            multiplicity += char * irrep.character(~element)
    
        multiplicity /= len(character) * irrep.sum_of_squares_constituents
    
        # the result has to be an integer
        assert math.isclose(multiplicity, round(multiplicity), abs_tol=1e-9), \
            "Multiplicity of irrep [%s] is not an integer: %f" % (str(irrep.id), multiplicity)
        # print(irrep_name, multiplicity)

        multiplicity = int(round(multiplicity))
        irreps += [irrep]*multiplicity
        multiplicities += [(irrep, multiplicity)]
    
    P = directsum(irreps, name="irreps")
    
    v = np.zeros((size, 1), dtype=float)
    
    p = 0
    for irr, m in multiplicities:
        assert irr.size >= m
        s = irr.size
        v[p:p+m*s, 0] = np.eye(m, s).reshape(-1) * np.sqrt(s)
        p += m*s
        
    change_of_basis = np.zeros((size, size))
    
    # np.set_printoptions(precision=4, threshold=10*size**2, suppress=False, linewidth=25*size + 5)
    
    for e in group.elements:
        ev = P(e) @ v
        change_of_basis[index[e], :] = ev.T
    
    change_of_basis /= np.sqrt(size)
    
    # the computed change of basis is an orthonormal matrix
    
    # change_of_basis_inv = sp.linalg.inv(change_of_basis)
    change_of_basis_inv = change_of_basis.T
    
    return irreps, change_of_basis, change_of_basis_inv
    
    # return Representation(group,
    #                       "regular",
    #                       [r.name for r in irreps],
    #                       change_of_basis,
    #                       ['pointwise', 'norm', 'gated', 'concatenated'],
    #                       representation=representation,
    #                       change_of_basis_inv=change_of_basis_inv)


def build_quotient_representation(group: escnn.group.Group,
                                  subgroup_id
                                  ) -> Tuple[List[escnn.group.IrreducibleRepresentation], np.ndarray, np.ndarray]:
    r"""

    Build the quotient representation of the input ``group`` with respect to the subgroup identified by ``subgroup_id``.
    
    .. seealso::
        See the :class:`~escnn.group.Group` instance's implementation of the method :meth:`~escnn.group.Group.subgroup`
        for more details on ``subgroup_id``.
    
    .. warning ::
        Only finite groups are supported
    
    Args:
        group (Group): the group whose representation has to be built
        subgroup_id: identifier of the subgroup

    Returns:
        a tuple containing the list of irreps, the change of basis and the inverse change of basis of
        the quotient representation

    """
    subgroup, _, _ = group.subgroup(subgroup_id)
    return build_induced_representation(group,
                                        subgroup_id,
                                        subgroup.trivial_representation)


def build_induced_representation(group: escnn.group.Group,
                                 subgroup_id,
                                 repr: escnn.group.IrreducibleRepresentation,
                                 representatives: List[GroupElement] = None,
                                 ) -> Tuple[List[escnn.group.IrreducibleRepresentation], np.ndarray, np.ndarray]:
    r"""

    Build the induced representation of the input ``group`` from the representation ``repr`` of the subgroup
    identified by ``subgroup_id``.

    .. seealso::
        See the :class:`~escnn.group.Group` instance's implementation of the method :meth:`~escnn.group.Group.subgroup`
        for more details on ``subgroup_id``.

    .. warning ::
        Only irreducible representations are supported as the subgroup representation.
        
    .. warning ::
        It is not possible to compute the index of [G:H] when they are not finite groups.
        Therefore, it is not possible to check that `representatives` contains sufficient elements.
        In case it does not, the construction of the final representation will probably fail.

    # TODO add note about the fact all irreps of `group` which contain `repr` need to be inside `group.irreps`
    
    Args:
        group (Group): the group whose representation has to be built
        subgroup_id: identifier of the subgroup
        repr (IrreducibleRepresentation): the representation of the subgroup
        representatives (list, optional): list of coset representatives used to define the induced representation

    Returns:
        a tuple containing the list of irreps, the change of basis and the inverse change of basis of
        the induced representation

    """
    
    assert repr.irreducible, "Induction from general representations is not supported yet"
    
    subgroup, parent, child = group.subgroup(subgroup_id)
    
    assert repr.group == subgroup
    homspace = group.homspace(subgroup_id)

    if group.order() > 0:
        quotient_size = int(group.order() / subgroup.order())
        
        if representatives is None:
            representatives = []
            # the coset each element belongs to
            cosets = {}
            for e in group.elements:
                if e not in cosets:
                    representatives.append(e)
                    for g in subgroup.elements:
                        eg = e @ homspace._inclusion(g)
                        cosets[eg] = e
                        
        assert quotient_size == len(representatives)
        
    else:
        assert representatives is not None
        quotient_size = len(representatives)
            
    size = repr.size * quotient_size
    
    # check that all representatives belong to different cosets
    for i, r1 in enumerate(representatives):
        for j, r2 in enumerate(representatives):
            if i != j:
                assert not homspace.same_coset(r1, r2), (r1, r2)

    change_of_basis = []
    irreps = []
    
    for rho in group.irreps():
        basis_rho = homspace._dirac_kernel_ft(rho.id, repr.id)
    
        m_rho = basis_rho.shape[1]
    
        irreps += [rho] * m_rho
    
        if m_rho > 0:
            ift = np.zeros((len(representatives), repr.size, m_rho, rho.size))
            for i, r in enumerate(representatives):
                ift[i, ...] = np.einsum('oi, imp->pmo', rho(r), basis_rho)
        
            ift /= np.sqrt(len(representatives))
            ift /= np.sqrt(repr.size)
            ift *= np.sqrt(rho.size)
        
            assert np.allclose(
                    ift.reshape(size, -1).T @ ift.reshape(size, -1),
                    np.eye(rho.size * m_rho)
            ), ift.reshape(size, -1).T @ ift.reshape(size, -1)
        
            ift = ift.reshape(size, rho.size * m_rho)
            change_of_basis.append(ift)

    change_of_basis = np.concatenate(change_of_basis, axis=1)
    
    assert change_of_basis.shape[0] <= change_of_basis.shape[1], f'Error! Induced representation not complete because group {group} is missing some irreps'
    assert change_of_basis.shape[0] >= change_of_basis.shape[1], f'Error! Induced representation not complete because the set of representatives passed is not complete'
    
    assert np.allclose(change_of_basis @ change_of_basis.T, np.eye(size))
    assert np.allclose(change_of_basis.T @ change_of_basis, np.eye(size))

    return irreps, change_of_basis, change_of_basis.T


def direct_sum_factory(irreps: List[escnn.group.IrreducibleRepresentation],
                       change_of_basis: np.ndarray,
                       change_of_basis_inv: np.ndarray = None
                       ) -> Callable[[GroupElement], np.ndarray]:
    """
    The method builds and returns a function implementing the direct sum of the "irreps" transformed by the given
    "change_of_basis" matrix.

    More precisely, the built method will take as input a value accepted by all the irreps, evaluate the irreps on that
    input and return the direct sum of the produced matrices left and right multiplied respectively by the
    change_of_basis matrix and its inverse.

    Args:
        irreps (list): list of irreps
        change_of_basis: the matrix transforming the direct sum of the irreps
        change_of_basis_inv: the inverse of the change of basis matrix

    Returns:
        function taking an input accepted by the irreps and returning the direct sum of the irreps evaluated
        on that input
    """
    size = sum(irr.size for irr in irreps)

    if change_of_basis is not None:
        shape = change_of_basis.shape
        assert len(shape) == 2 and shape[0] == shape[1]
        assert shape[0] == size

        if change_of_basis_inv is None:
            # pre-compute the inverse of the change-of-_bases matrix
            change_of_basis_inv = linalg.inv(change_of_basis)
        else:
            assert len(change_of_basis_inv.shape) == 2
            assert change_of_basis_inv.shape[0] == change_of_basis.shape[0]
            assert change_of_basis_inv.shape[1] == change_of_basis.shape[1]
            assert np.allclose(change_of_basis @ change_of_basis_inv, np.eye(change_of_basis.shape[0]))
            assert np.allclose(change_of_basis_inv @ change_of_basis, np.eye(change_of_basis.shape[0]))
    
    unique_irreps = list({irr.id: irr for irr in irreps}.items())
    irreps_ids = [irr.id for irr in irreps]
    
    def direct_sum(element: GroupElement,
                   irreps_ids=irreps_ids, change_of_basis=change_of_basis,
                   change_of_basis_inv=change_of_basis_inv, unique_irreps=unique_irreps):
        reprs = {}
        for n, irr in unique_irreps:
            reprs[n] = irr(element)
        
        blocks = []
        for irrep_id in irreps_ids:
            repr = reprs[irrep_id]
            blocks.append(repr)
        
        P = sparse.block_diag(blocks, format='csc')

        if change_of_basis is None:
            return np.asarray(P.todense())
        else:
            return change_of_basis @ P @ change_of_basis_inv
    
    return direct_sum


def homomorphism_space(rho1: Representation, rho2: Representation) -> np.ndarray:
    r"""
        Compute a basis for the space of homomorphisms from `rho1` to `rho2`.

        .. warning::
            This basis might become quite big for large representations and, therefore, this method might get
            computationally expensive.
            In practice, we recommend using this method only to generate the basis for small representations (e.g.
            for irreps).
            This basis is generated and stored more efficiently in the :doc:`escnn.nn` package, when parameterizing
            neural network modules; e.g. see :class:`~escnn.nn.BlocksBasisExpansion`.
            
    """
    assert rho1.group == rho2.group
    G = rho1.group
    
    dim = 0
    for in_irrep in rho1.irreps:
        in_irrep = G.irrep(*in_irrep)
        for out_irrep in rho2.irreps:
            out_irrep = G.irrep(*out_irrep)
            if in_irrep == out_irrep:
                dim += in_irrep.sum_of_squares_constituents
    
    basis = np.zeros((dim, rho2.size, rho1.size))
    
    p = 0
    i = 0
    for in_irrep in rho1.irreps:
        in_irrep = G.irrep(*in_irrep)
        
        j = 0
        for out_irrep in rho2.irreps:
            out_irrep = G.irrep(*out_irrep)
            
            if out_irrep == in_irrep:
                basis[
                    p:p + in_irrep.sum_of_squares_constituents,
                    j:j + in_irrep.size,
                    i:i + in_irrep.size
                ] = in_irrep.endomorphism_basis()
                
                p += in_irrep.sum_of_squares_constituents
            
            j += out_irrep.size
        i += in_irrep.size
    
    basis = np.einsum('Mm,kmn,nN->kMN', rho2.change_of_basis, basis, rho1.change_of_basis_inv)
    return basis





from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch
import torch.nn.functional as F

from typing import List, Tuple, Any, Union

__all__ = [
    "PointwiseAdaptiveMaxPool2D", "PointwiseAdaptiveMaxPool",
    "PointwiseAdaptiveMaxPool3D",
]


class _PointwiseAdaptiveMaxPoolND(EquivariantModule):
    
    def __init__(self,
                 in_type: FieldType,
                 d: int,
                 output_size: Union[int, Tuple]
                 ):
        r"""

        Module that implements adaptive channel-wise max-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.AdaptiveMaxPool2D` or :class:`torch.nn.AdaptiveMaxPool3D`,
        wrapping it in the :class:`~escnn.nn.EquivariantModule` interface.
        
        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.
            
        Args:
            in_type (FieldType): the input field type
            d (int): dimensionality of the base space (2 for images, 3 for volumes)
            output_size: the target output size

        """

        assert d in [2, 3], f"Only dimensionality 2 or 3 are currently suported by 'd={d}' found"

        assert isinstance(in_type.gspace, GSpace)
        assert in_type.gspace.dimensionality == d, (in_type.gspace.dimensionality, d)

        for r in in_type.representations:
            assert 'pointwise' in r.supported_nonlinearities, \
                f"""Error! Representation "{r}" does not support pointwise non-linearities
                so it is not possible to pool each channel independently"""
        
        super(_PointwiseAdaptiveMaxPoolND, self).__init__()

        self.d = d
        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type

        if isinstance(output_size, int):
            self.output_size = (output_size,) * self.d
        else:
            self.output_size = output_size

        assert isinstance(self.output_size, tuple) and len(self.output_size) == self.d, self.output_size

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert input.type == self.in_type
        
        # run the common max-pooling
        if self.d == 2:
            output = F.adaptive_max_pool2d(input.tensor, self.output_size)
        elif self.d == 3:
            output = F.adaptive_max_pool3d(input.tensor, self.output_size)
        else:
            raise NotImplementedError

        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords=None)

    def evaluate_output_shape(self, input_shape: Tuple) -> Tuple:
        assert len(input_shape) == 2 + self.d
        assert input_shape[1] == self.in_type.size
        return (input_shape[0], self.out_type.size, *self.output_size)

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
    
        # this kind of pooling is not really equivariant so we can not test equivariance
        pass

    def export(self):
        r"""
        Export this module to a normal PyTorch :class:`torch.nn.AdaptiveMaxPool2d` or
        :class:`torch.nn.AdaptiveMaxPool3d` module and set to "eval" mode.

        """
    
        self.eval()

        if self.d == 2:
            return torch.nn.AdaptiveMaxPool2d(self.output_size).eval()
        elif self.d == 3:
            return torch.nn.AdaptiveMaxPool3d(self.output_size).eval()
        else:
            raise NotImplementedError


class PointwiseAdaptiveMaxPool2D(_PointwiseAdaptiveMaxPoolND):

    def __init__(self,
                 in_type: FieldType,
                 output_size: Union[int, Tuple[int, int]]
                 ):
        r"""

        Module that implements adaptive channel-wise max-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.AdaptiveMaxPool2D`, wrapping it in the
        :class:`~escnn.nn.EquivariantModule` interface.

        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            output_size: the target output size of the image of the form H x W

        """

        assert isinstance(in_type.gspace, GSpace)
        assert in_type.gspace.dimensionality == 2

        super(PointwiseAdaptiveMaxPool2D, self).__init__(
            in_type, 2, output_size
        )


class PointwiseAdaptiveMaxPool3D(_PointwiseAdaptiveMaxPoolND):

    def __init__(self,
                 in_type: FieldType,
                 output_size: Union[int, Tuple[int, int]]
                 ):
        r"""

        Module that implements adaptive channel-wise max-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.AdaptiveMaxPool3D`, wrapping it in the
        :class:`~escnn.nn.EquivariantModule` interface.

        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            output_size: the target output size of the image of the form H x W

        """

        assert isinstance(in_type.gspace, GSpace)
        assert in_type.gspace.dimensionality == 3

        super(PointwiseAdaptiveMaxPool3D, self).__init__(
            in_type, 3, output_size
        )


# for backward compatibility
PointwiseAdaptiveMaxPool = PointwiseAdaptiveMaxPool2D



from escnn.nn import FieldType
from typing import List, Dict, Tuple
from collections import defaultdict


__all__ = ["check_consecutive_numbers", "indexes_from_labels"]


def check_consecutive_numbers(list: List[int]) -> bool:
    
    m = M = list[0]
    s = 0
    
    for l in list:
        assert l >= 0
        m = min(m, l)
        M = max(M, l)
        s += l
    
    S = M*(M+1)/2 - (m-1)*m/2
    
    return S == s
    
    
def indexes_from_labels(in_type: FieldType, labels: List[str]) -> Dict[str, Tuple[bool, List[int], List[int]]]:
    r"""
    
    
    Args:
        in_type (FieldType): the input field type
        labels (list): a list of strings long as the list :attr:'~escnn.nn.FieldType.representations`
                of the input :attr:`in_type`

    Returns:

    """
    assert len(labels) == len(in_type)
    
    indeces = defaultdict(lambda: [])
    fields = defaultdict(lambda: [])
    
    current_position = 0
    for c, (l, r) in enumerate(zip(labels, in_type.representations)):
        # append the indeces of the current field to the list corresponding to this label
        indeces[l] += list(range(current_position, current_position + r.size))
        fields[l].append(c)
        current_position += r.size
    
    groups = {}
    
    for l in labels:
        contiguous = check_consecutive_numbers(indeces[l])
        groups[l] = contiguous, fields[l], indeces[l]
    
    return groups

